# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tacker._i18n import _
from tacker.api.v1 import attributes
from tacker.common import exceptions as n_exc
from tacker.tests import base


class TestAttributes(base.BaseTestCase):

    def _construct_dict_and_constraints(self):
        """Constructs a test dictionary and a definition of constraints.

        :return: A (dictionary, constraint) tuple
        """
        constraints = {'key1': {'type:values': ['val1', 'val2'],
                                'required': True},
                       'key2': {'type:string': None,
                                'required': False},
                       'key3': {'type:dict': {'k4': {'type:string': None,
                                                     'required': True}},
                                'required': True}}

        dictionary = {'key1': 'val1',
                      'key2': 'a string value',
                      'key3': {'k4': 'a string value'}}

        return dictionary, constraints

    def test_is_attr_set(self):
        data = attributes.ATTR_NOT_SPECIFIED
        self.assertFalse(attributes.is_attr_set(data))

        data = None
        self.assertFalse(attributes.is_attr_set(data))

        data = "I'm set"
        self.assertTrue(attributes.is_attr_set(data))

    def test_validate_values(self):
        msg = attributes._validate_values(4, [4, 6])
        self.assertIsNone(msg)

        msg = attributes._validate_values(4, (4, 6))
        self.assertIsNone(msg)

        msg = attributes._validate_values(7, [4, 6])
        self.assertNotIn(7, [4, 6], msg)

        msg = attributes._validate_values(7, (4, 6))
        self.assertNotIn(7, (4, 6), msg)

    def test_validate_not_empty_string(self):
        msg = attributes._validate_not_empty_string('    ', None)
        self.assertEqual("'    ' Blank strings are not permitted", msg)

    def test_validate_not_empty_string_or_none(self):
        msg = attributes._validate_not_empty_string_or_none('    ', None)
        self.assertEqual("'    ' Blank strings are not permitted", msg)

        msg = attributes._validate_not_empty_string_or_none(None, None)
        self.assertIsNone(msg)

    def test_validate_string_or_none(self):
        msg = attributes._validate_not_empty_string_or_none('test', None)
        self.assertIsNone(msg)

        msg = attributes._validate_not_empty_string_or_none(None, None)
        self.assertIsNone(msg)

    def test_validate_string(self):
        msg = attributes._validate_string(None, None)
        self.assertEqual("'None' is not a valid string", msg)

        # 0 == len(data) == max_len
        msg = attributes._validate_string("", 0)
        self.assertIsNone(msg)

        # 0 == len(data) < max_len
        msg = attributes._validate_string("", 9)
        self.assertIsNone(msg)

        # 0 < len(data) < max_len
        msg = attributes._validate_string("123456789", 10)
        self.assertIsNone(msg)

        # 0 < len(data) == max_len
        msg = attributes._validate_string("123456789", 9)
        self.assertIsNone(msg)

        # 0 < max_len < len(data)
        msg = attributes._validate_string("1234567890", 9)
        self.assertEqual("'1234567890' exceeds maximum length of 9", msg)

        msg = attributes._validate_string("123456789", None)
        self.assertIsNone(msg)

    def test_validate_no_whitespace(self):
        data = 'no_white_space'
        result = attributes._validate_no_whitespace(data)
        self.assertEqual(data, result)

        self.assertRaises(n_exc.InvalidInput,
                          attributes._validate_no_whitespace,
                          'i have whitespace')

        self.assertRaises(n_exc.InvalidInput,
                          attributes._validate_no_whitespace,
                          'i\thave\twhitespace')

    def test_validate_range(self):
        msg = attributes._validate_range(1, [1, 9])
        self.assertIsNone(msg)

        msg = attributes._validate_range(5, [1, 9])
        self.assertIsNone(msg)

        msg = attributes._validate_range(9, [1, 9])
        self.assertIsNone(msg)

        msg = attributes._validate_range(1, (1, 9))
        self.assertIsNone(msg)

        msg = attributes._validate_range(5, (1, 9))
        self.assertIsNone(msg)

        msg = attributes._validate_range(9, (1, 9))
        self.assertIsNone(msg)

        msg = attributes._validate_range(0, [1, 9])
        self.assertEqual("'0' is too small - must be at least '1'", msg)

        msg = attributes._validate_range(10, (1, 9))
        self.assertEqual("'10' is too large - must be no larger than '9'",
                         msg)

        msg = attributes._validate_range("bogus", (1, 9))
        self.assertEqual("'bogus' is not an integer", msg)

        msg = attributes._validate_range(10, (attributes.UNLIMITED,
                                              attributes.UNLIMITED))
        self.assertIsNone(msg)

        msg = attributes._validate_range(10, (1, attributes.UNLIMITED))
        self.assertIsNone(msg)

        msg = attributes._validate_range(1, (attributes.UNLIMITED, 9))
        self.assertIsNone(msg)

        msg = attributes._validate_range(-1, (0, attributes.UNLIMITED))
        self.assertEqual("'-1' is too small - must be at least '0'", msg)

        msg = attributes._validate_range(10, (attributes.UNLIMITED, 9))
        self.assertEqual("'10' is too large - must be no larger than '9'", msg)

    def _test_validate_mac_address(self, validator, allow_none=False):
        mac_addr = "ff:16:3e:4f:00:00"
        msg = validator(mac_addr)
        self.assertIsNone(msg)

        mac_addr = "ffa:16:3e:4f:00:00"
        msg = validator(mac_addr)
        err_msg = "'%s' is not a valid MAC address"
        self.assertEqual(err_msg % mac_addr, msg)

        mac_addr = "123"
        msg = validator(mac_addr)
        self.assertEqual(err_msg % mac_addr, msg)

        mac_addr = None
        err_msg = "MAC address must be string: 'NoneType' object has" \
                  " no attribute 'split'"
        msg = validator(mac_addr)
        if allow_none:
            self.assertIsNone(msg)
        else:
            self.assertEqual(err_msg, msg)

    def test_validate_mac_address(self):
        self._test_validate_mac_address(attributes._validate_mac_address)

    def test_validate_mac_address_or_none(self):
        self._test_validate_mac_address(
            attributes._validate_mac_address_or_none, allow_none=True)

    def test_validate_ip_address(self):
        ip_addr = '1.1.1.1'
        msg = attributes._validate_ip_address(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1111.1.1.1'
        msg = attributes._validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '1.1.1.1 has whitespace'
        msg = attributes._validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '111.1.1.1\twhitespace'
        msg = attributes._validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '111.1.1.1\nwhitespace'
        msg = attributes._validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

    def test_validate_ip_pools(self):
        pools = [[{'end': '10.0.0.254'}],
                 [{'start': '10.0.0.254'}],
                 [{'start': '1000.0.0.254',
                   'end': '1.1.1.1'}],
                 [{'start': '10.0.0.2', 'end': '10.0.0.254',
                   'forza': 'juve'}],
                 [{'start': '10.0.0.2', 'end': '10.0.0.254'},
                  {'end': '10.0.0.254'}],
                 [None],
                 None]
        for pool in pools:
            msg = attributes._validate_ip_pools(pool)
            self.assertIsNotNone(msg)

        pools = [[{'end': '10.0.0.254', 'start': '10.0.0.2'},
                  {'start': '11.0.0.2', 'end': '11.1.1.1'}],
                 [{'start': '11.0.0.2', 'end': '11.0.0.100'}]]
        for pool in pools:
            msg = attributes._validate_ip_pools(pool)
            self.assertIsNone(msg)

    def test_validate_fixed_ips(self):
        fixed_ips = [
            {'data': [{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1111.1.1.1'}],
             'error_msg': "'1111.1.1.1' is not a valid IP address"},
            {'data': [{'subnet_id': 'invalid',
                       'ip_address': '1.1.1.1'}],
             'error_msg': "'invalid' is not a valid UUID"},
            {'data': None,
             'error_msg': "Invalid data format for fixed IP: 'None'"},
            {'data': "1.1.1.1",
             'error_msg': "Invalid data format for fixed IP: '1.1.1.1'"},
            {'data': ['00000000-ffff-ffff-ffff-000000000000', '1.1.1.1'],
             'error_msg': "Invalid data format for fixed IP: "
                          "'00000000-ffff-ffff-ffff-000000000000'"},
            {'data': [['00000000-ffff-ffff-ffff-000000000000', '1.1.1.1']],
             'error_msg': "Invalid data format for fixed IP: "
                          "'['00000000-ffff-ffff-ffff-000000000000', "
                          "'1.1.1.1']'"},
            {'data': [{'subnet_id': '00000000-0fff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'},
                      {'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'}],
             'error_msg': "Duplicate IP address '1.1.1.1'"}]
        for fixed in fixed_ips:
            msg = attributes._validate_fixed_ips(fixed['data'])
            self.assertEqual(fixed['error_msg'], msg)

        fixed_ips = [[{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'}],
                     [{'subnet_id': '00000000-0fff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'},
                      {'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.2'}]]
        for fixed in fixed_ips:
            msg = attributes._validate_fixed_ips(fixed)
            self.assertIsNone(msg)

    def test_validate_nameservers(self):
        ns_pools = [['1.1.1.2', '1.1.1.2'],
                    ['www.hostname.com', 'www.hostname.com'],
                    ['77.hostname.com'],
                    ['1000.0.0.1'],
                    None]

        for ns in ns_pools:
            msg = attributes._validate_nameservers(ns, None)
            self.assertIsNotNone(msg)

        ns_pools = [['100.0.0.2'],
                    ['www.hostname.com'],
                    ['www.great.marathons.to.travel'],
                    ['valid'],
                    ['www.internal.hostname.com']]

        for ns in ns_pools:
            msg = attributes._validate_nameservers(ns, None)
            self.assertIsNone(msg)

    def test_validate_hostroutes(self):
        hostroute_pools = [[{'destination': '100.0.0.0/24'}],
                           [{'nexthop': '10.0.2.20'}],
                           [{'nexthop': '10.0.2.20',
                             'forza': 'juve',
                             'destination': '100.0.0.0/8'}],
                           [{'nexthop': '1110.0.2.20',
                             'destination': '100.0.0.0/8'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'},
                            {'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'}],
                           [None],
                           None]
        for host_routes in hostroute_pools:
            msg = attributes._validate_hostroutes(host_routes, None)
            self.assertIsNotNone(msg)

        hostroute_pools = [[{'destination': '100.0.0.0/24',
                             'nexthop': '10.0.2.20'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'},
                            {'nexthop': '10.0.2.20',
                             'destination': '101.0.0.0/8'}]]
        for host_routes in hostroute_pools:
            msg = attributes._validate_hostroutes(host_routes, None)
            self.assertIsNone(msg)

    def test_validate_ip_address_or_none(self):
        ip_addr = None
        msg = attributes._validate_ip_address_or_none(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1.1.1.1'
        msg = attributes._validate_ip_address_or_none(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1111.1.1.1'
        msg = attributes._validate_ip_address_or_none(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr,
                         msg)

    def test_hostname_pattern(self):
        data = '@openstack'
        msg = attributes._validate_regex(data, attributes.HOSTNAME_PATTERN)
        self.assertIsNotNone(msg)

        data = 'www.openstack.org'
        msg = attributes._validate_regex(data, attributes.HOSTNAME_PATTERN)
        self.assertIsNone(msg)

    def test_uuid_pattern(self):
        data = 'garbage'
        msg = attributes._validate_regex(data, attributes.UUID_PATTERN)
        self.assertIsNotNone(msg)

        data = '00000000-ffff-ffff-ffff-000000000000'
        msg = attributes._validate_regex(data, attributes.UUID_PATTERN)
        self.assertIsNone(msg)

    def test_mac_pattern(self):
        # Valid - 3 octets
        base_mac = "fa:16:3e:00:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNone(msg)

        # Valid - 4 octets
        base_mac = "fa:16:3e:4f:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNone(msg)

        # Invalid - not unicast
        base_mac = "01:16:3e:4f:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "a:16:3e:4f:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "ffa:16:3e:4f:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "01163e4f0000"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "01-16-3e-4f-00-00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "00:16:3:f:00:00"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "12:3:4:5:67:89ab"
        msg = attributes._validate_regex(base_mac,
                                         attributes.MAC_PATTERN)
        self.assertIsNotNone(msg)

    def _test_validate_subnet(self, validator, allow_none=False):
        # Valid - IPv4
        cidr = "10.0.2.0/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - IPv6 without final octets
        cidr = "fe80::/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - IPv6 with final octets
        cidr = "fe80::/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - uncompressed ipv6 address
        cidr = "fe80:0:0:0:0:0:0:0/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001:0db8:0:0:1::1/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001:0db8::1:0:0:1/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001::0:1:0:0:1100/120"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - abbreviated ipv4 address
        cidr = "10/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Invalid - IPv4 missing mask
        cidr = "10.0.2.0"
        msg = validator(cidr, None)
        error = _("'%(data)s' isn't a recognized IP subnet cidr,"
                  " '%(cidr)s' is recommended") % {"data": cidr,
                                                   "cidr": "10.0.2.0/32"}
        self.assertEqual(error, msg)

        # Valid - IPv4 with non-zero masked bits is ok
        for i in range(1, 255):
            cidr = "192.168.1.%s/24" % i
            msg = validator(cidr, None)
            self.assertIsNone(msg)

        # Invalid - IPv6 without final octets, missing mask
        cidr = "fe80::"
        msg = validator(cidr, None)
        error = _("'%(data)s' isn't a recognized IP subnet cidr,"
                  " '%(cidr)s' is recommended") % {"data": cidr,
                                                   "cidr": "fe80::/128"}
        self.assertEqual(error, msg)

        # Invalid - IPv6 with final octets, missing mask
        cidr = "fe80::0"
        msg = validator(cidr, None)
        error = _("'%(data)s' isn't a recognized IP subnet cidr,"
                  " '%(cidr)s' is recommended") % {"data": cidr,
                                                   "cidr": "fe80::/128"}
        self.assertEqual(error, msg)

        # Invalid - Address format error
        cidr = 'invalid'
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        cidr = None
        msg = validator(cidr, None)
        if allow_none:
            self.assertIsNone(msg)
        else:
            error = "'%s' is not a valid IP subnet" % cidr
            self.assertEqual(error, msg)

    def test_validate_subnet(self):
        self._test_validate_subnet(attributes._validate_subnet)

    def test_validate_subnet_or_none(self):
        self._test_validate_subnet(attributes._validate_subnet_or_none,
                                   allow_none=True)

    def test_validate_subnet_list(self):
        subnet_list = ['10.0.2.0/24', 'fe80::/24']
        result = attributes._validate_subnet_list(subnet_list)
        self.assertEqual(None, result)

        subnet_list = ['10.0.2.0/24', 'fe80::/24', '10.0.2.0/24']
        expect_res = "Duplicate items in the list: '10.0.2.0/24, " \
                     "fe80::/24, 10.0.2.0/24'"
        result = attributes._validate_subnet_list(subnet_list)
        self.assertEqual(expect_res, result)

    def _test_validate_regex(self, validator, allow_none=False):
        pattern = '[hc]at'

        data = None
        msg = validator(data, pattern)
        if allow_none:
            self.assertIsNone(msg)
        else:
            self.assertEqual("'None' is not a valid input", msg)

        data = 'bat'
        msg = validator(data, pattern)
        self.assertEqual("'%s' is not a valid input" % data, msg)

        data = 'hat'
        msg = validator(data, pattern)
        self.assertIsNone(msg)

        data = 'cat'
        msg = validator(data, pattern)
        self.assertIsNone(msg)

    def test_validate_regex(self):
        self._test_validate_regex(attributes._validate_regex)

    def test_validate_regex_or_none(self):
        self._test_validate_regex(attributes._validate_regex_or_none,
                                  allow_none=True)

    def test_validate_uuid(self):
        msg = attributes._validate_uuid('garbage')
        self.assertEqual("'garbage' is not a valid UUID", msg)

        msg = attributes._validate_uuid('00000000-ffff-ffff-ffff-000000000000')
        self.assertIsNone(msg)

    def test_validate_uuid_list(self):
        # check not a list
        uuids = [None,
                 123,
                 'e5069610-744b-42a7-8bd8-ceac1a229cd4',
                 '12345678123456781234567812345678',
                 {'uuid': 'e5069610-744b-42a7-8bd8-ceac1a229cd4'}]
        for uuid in uuids:
            msg = attributes._validate_uuid_list(uuid)
            error = "'%s' is not a list" % uuid
            self.assertEqual(error, msg)

        # check invalid uuid in a list
        invalid_uuid_lists = [[None],
                              [123],
                              [123, 'e5069610-744b-42a7-8bd8-ceac1a229cd4'],
                              ['123', '12345678123456781234567812345678'],
                              ['t5069610-744b-42a7-8bd8-ceac1a229cd4'],
                              ['e5069610-744b-42a7-8bd8-ceac1a229cd44'],
                              ['e50696100-744b-42a7-8bd8-ceac1a229cd4'],
                              ['e5069610-744bb-42a7-8bd8-ceac1a229cd4']]
        for uuid_list in invalid_uuid_lists:
            msg = attributes._validate_uuid_list(uuid_list)
            error = "'%s' is not a valid UUID" % uuid_list[0]
            self.assertEqual(error, msg)

        # check duplicate items in a list
        duplicate_uuids = ['e5069610-744b-42a7-8bd8-ceac1a229cd4',
                           'f3eeab00-8367-4524-b662-55e64d4cacb5',
                           'e5069610-744b-42a7-8bd8-ceac1a229cd4']
        msg = attributes._validate_uuid_list(duplicate_uuids)
        error = ("Duplicate items in the list: "
                 "'%s'" % ', '.join(duplicate_uuids))
        self.assertEqual(error, msg)

        # check valid uuid lists
        valid_uuid_lists = [['e5069610-744b-42a7-8bd8-ceac1a229cd4'],
                            ['f3eeab00-8367-4524-b662-55e64d4cacb5'],
                            ['e5069610-744b-42a7-8bd8-ceac1a229cd4',
                             'f3eeab00-8367-4524-b662-55e64d4cacb5']]
        for uuid_list in valid_uuid_lists:
            msg = attributes._validate_uuid_list(uuid_list)
            self.assertIsNone(msg)

    def test_validate_dict_type(self):
        for value in (None, True, '1', []):
            self.assertEqual("'%s' is not a dictionary" % value,
                             attributes._validate_dict(value))

    def test_validate_dict_without_constraints(self):
        msg = attributes._validate_dict({})
        self.assertIsNone(msg)

        # Validate a dictionary without constraints.
        msg = attributes._validate_dict({'key': 'value'})
        self.assertIsNone(msg)

    def test_validate_a_valid_dict_with_constraints(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_with_invalid_validator(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        constraints['key1'] = {'type:unsupported': None, 'required': True}
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertEqual("Validator 'type:unsupported' does not exist.",
                         msg)

    def test_validate_dict_not_required_keys(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key2']
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIsNone(msg, 'Field that was not required by the specs was '
                               'required by the validator.')

    def test_validate_dict_required_keys(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key1']
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIn('Expected keys:', msg)

    def test_validate_dict_wrong_values(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        dictionary['key1'] = 'UNSUPPORTED'
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIsNotNone(msg)

    def test_validate_dict_convert_boolean(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        constraints['key_bool'] = {
            'type:boolean': None,
            'required': False,
            'convert_to': attributes.convert_to_boolean}
        dictionary['key_bool'] = 'true'
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIsNone(msg)
        # Explicitly comparing with literal 'True' as assertTrue
        # succeeds also for 'true'
        self.assertTrue(dictionary['key_bool'])

    def test_subdictionary(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key3']['k4']
        dictionary['key3']['k5'] = 'a string value'
        msg = attributes._validate_dict(dictionary, constraints)
        self.assertIn('Expected keys:', msg)

    def test_validate_dict_or_none(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        # Check whether None is a valid value.
        msg = attributes._validate_dict_or_none(None, constraints)
        self.assertIsNone(msg, 'Validation of a None dictionary failed.')

        # Check validation of a regular dictionary.
        msg = attributes._validate_dict_or_none(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_or_empty(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        # Check whether an empty dictionary is valid.
        msg = attributes._validate_dict_or_empty({}, constraints)
        self.assertIsNone(msg, 'Validation of a None dictionary failed.')

        # Check validation of a regular dictionary.
        msg = attributes._validate_dict_or_none(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_or_not_empty(self):
        dictionary, constraints = self._construct_dict_and_constraints()
        msg = attributes._validate_dict_not_empty({}, constraints)
        self.assertEqual(msg, "'{}' is not a valid dictionary or it is"
                              " an empty dictionary")
        msg = attributes._validate_dict_not_empty("", constraints)
        self.assertEqual(msg, "'' is not a valid dictionary or it is an"
                              " empty dictionary")

    def test_validate_non_negative(self):
        for value in (-1, '-2'):
            self.assertEqual("'%s' should be non-negative" % value,
                             attributes._validate_non_negative(value))

        for value in (0, 1, '2', True, False):
            msg = attributes._validate_non_negative(value)
            self.assertIsNone(msg)


class TestConvertToBoolean(base.BaseTestCase):

    def test_convert_to_boolean_bool(self):
        self.assertTrue(attributes.convert_to_boolean(True))
        self.assertFalse(attributes.convert_to_boolean(False))

    def test_convert_to_boolean_int(self):
        self.assertFalse(attributes.convert_to_boolean(0))
        self.assertTrue(attributes.convert_to_boolean(1))
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_boolean,
                          7)

    def test_convert_to_boolean_str(self):
        self.assertTrue(attributes.convert_to_boolean('True'))
        self.assertTrue(attributes.convert_to_boolean('true'))
        self.assertFalse(attributes.convert_to_boolean('False'))
        self.assertFalse(attributes.convert_to_boolean('false'))
        self.assertFalse(attributes.convert_to_boolean('0'))
        self.assertTrue(attributes.convert_to_boolean('1'))
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_boolean,
                          '7')
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_boolean,
                          "")
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_boolean,
                          "abc")
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_boolean,
                          123)


class TestConvertToInt(base.BaseTestCase):

    def test_convert_to_int_int(self):
        self.assertEqual(-1, attributes.convert_to_int(-1))
        self.assertEqual(0, attributes.convert_to_int(0))
        self.assertEqual(1, attributes.convert_to_int(1))

    def test_convert_to_int_str(self):
        self.assertEqual(4, attributes.convert_to_int('4'))
        self.assertEqual(6, attributes.convert_to_int('6'))
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_int,
                          'garbage')

    def test_convert_to_int_none(self):
        self.assertRaises(n_exc.InvalidInput,
                          attributes.convert_to_int,
                          None)

    def test_convert_none_to_empty_list_none(self):
        self.assertEqual(
            [], attributes.convert_none_to_empty_list(None))

    def test_convert_none_to_empty_dict(self):
        self.assertEqual(
            {}, attributes.convert_none_to_empty_dict(None))

    def test_convert_none_to_empty_list_value(self):
        values = ['1', 3, [], [1], {}, {'a': 3}]
        for value in values:
            self.assertEqual(
                value, attributes.convert_none_to_empty_list(value))


class TestConvertKvp(base.BaseTestCase):

    def test_convert_kvp_list_to_dict_succeeds_for_missing_values(self):
        result = attributes.convert_kvp_list_to_dict(['True'])
        self.assertEqual({}, result)

    def test_convert_kvp_list_to_dict_succeeds_for_multiple_values(self):
        self.skipTest("Not ready yet")
        result = attributes.convert_kvp_list_to_dict(
            ['a=b', 'a=c', 'a=c', 'b=a'])
        self.assertEqual({'a': ['c', 'b'], 'b': ['a']}, result)

    def test_convert_kvp_list_to_dict_succeeds_for_values(self):
        result = attributes.convert_kvp_list_to_dict(['a=b', 'c=d'])
        self.assertEqual({'a': ['b'], 'c': ['d']}, result)

    def test_convert_kvp_str_to_list_fails_for_missing_key(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            attributes.convert_kvp_str_to_list('=a')

    def test_convert_kvp_str_to_list_fails_for_missing_equals(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            attributes.convert_kvp_str_to_list('a')

    def test_convert_kvp_str_to_list_succeeds_for_one_equals(self):
        result = attributes.convert_kvp_str_to_list('a=')
        self.assertEqual(['a', ''], result)

    def test_convert_kvp_str_to_list_succeeds_for_two_equals(self):
        result = attributes.convert_kvp_str_to_list('a=a=a')
        self.assertEqual(['a', 'a=a'], result)


class TestConvertToList(base.BaseTestCase):

    def test_convert_to_empty_list(self):
        for item in (None, [], (), {}):
            self.assertEqual([], attributes.convert_to_list(item))

    def test_convert_to_list_string(self):
        for item in ('', 'foo'):
            self.assertEqual([item], attributes.convert_to_list(item))

    def test_convert_to_list_iterable(self):
        for item in ([None], [1, 2, 3], (1, 2, 3), set([1, 2, 3]), ['foo']):
            self.assertEqual(list(item), attributes.convert_to_list(item))

    def test_convert_to_list_non_iterable(self):
        for item in (True, False, 1, 1.2, object()):
            self.assertEqual([item], attributes.convert_to_list(item))



# Copyright (C) 2021 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import os
import time

import eventlet
from oslo_log import log as logging
import paramiko
import yaml

from tacker.common import cmd_executer
from tacker.common import exceptions
from tacker.common import log
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver

COMMAND_WAIT_RETRY_TIME = 30
CONNECT_REMOTE_SERVER_RETRY_COUNT = 4
LOG = logging.getLogger(__name__)
K8S_CMD_TIMEOUT = 60
K8S_DEPLOY_TIMEOUT = 300
SERVER_WAIT_COMPLETE_TIME = 120


class CnfNodePortMgmt(vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):
    def get_type(self):
        return 'mgmt-drivers-cnf-nodeport'

    def get_name(self):
        return 'mgmt-drivers-cnf-nodeport'

    def get_description(self):
        return 'Tacker CNFMgmt NodePort Setting Driver'

    @log.log
    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        pass

    def _check_is_cidr(self, key, value, cidr_str):
        # instantiate: check cidr
        try:
            ipaddress.ip_network(cidr_str)
        except ValueError:
            LOG.error('The {value} of {key} in the '
                      'additionalParams is invalid.'.format(
                          value=value, key=key))
            raise exceptions.MgmtDriverParamInvalid(param=value)

    def _check_input_parameters(self, additional_param, vnf_package_path):
        if not additional_param:
            LOG.error("The 'lcm-kubernetes-external-lb' cannot be None "
                      "in additionalParams.")
            raise exceptions.MgmtDriverOtherError(
                error_message="The 'lcm-kubernetes-external-lb' cannot"
                              " be None in additionalParams.")
        if not isinstance(additional_param, dict):
            LOG.error("The format of 'lcm-kubernetes-external-lb' in "
                      "additionalParams is invalid. It must be dict.")
            raise exceptions.MgmtDriverOtherError(
                error_message="The format of 'lcm-kubernetes-external-lb' in "
                              "additionalParams is invalid. It must be dict.")
        for key, value in additional_param.items():
            if key == 'external_lb_param':
                for attr in ['ssh_username', 'ssh_password', 'ssh_ip']:
                    if not value.get(attr):
                        LOG.error(
                            'The {} of {} in the '
                            'additionalParams cannot'
                            ' be None.'.format(attr, key))
                        raise exceptions.MgmtDriverNotFound(
                            param=attr)
                if value.get('ssh_ip'):
                    self._check_is_cidr(
                        key, 'ssh_ip', value.get('ssh_ip'))
        if not additional_param.get('script_path'):
            LOG.error('The script_path of {} in the '
                      'additionalParams cannot be None.'.format(key))
            raise exceptions.MgmtDriverNotFound(
                param='script_path')
        abs_script_path = os.path.join(
            vnf_package_path, additional_param.get('script_path'))
        if not os.path.exists(abs_script_path):
            LOG.error('The path of external_lb_param'
                      ' script is invalid.')
            raise exceptions.MgmtDriverOtherError(
                error_message="The path of external_lb_param"
                              " script is invalid")

    def _init_commander_and_set_script(self, user, password, host,
                                       timeout, vnf_package_path=None,
                                       script_path=None):
        retry = CONNECT_REMOTE_SERVER_RETRY_COUNT
        while retry > 0:
            try:
                if (vnf_package_path and script_path):
                    connect = paramiko.Transport(host, 22)
                    connect.connect(username=user, password=password)
                    sftp = paramiko.SFTPClient.from_transport(connect)
                    sftp.put(os.path.join(vnf_package_path, script_path),
                             "/tmp/{}".format(
                                 script_path.replace('Scripts', '')))
                    connect.close()
                commander = cmd_executer.RemoteCommandExecutor(
                    user=user, password=password, host=host,
                    timeout=timeout)
                return commander
            except (exceptions.NotAuthorized, paramiko.SSHException,
                    paramiko.ssh_exception.NoValidConnectionsError) as e:
                LOG.debug(e)
                retry -= 1
                if retry == 0:
                    LOG.error(e)
                    raise paramiko.SSHException()
                time.sleep(SERVER_WAIT_COMPLETE_TIME)

    def _execute_command(self, commander, ssh_command, timeout, type, retry):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    result = commander.execute_command(
                        ssh_command, input_data=None)
                    break
            except eventlet.timeout.Timeout:
                LOG.debug('It is time out, When execute command: '
                          '{}.'.format(ssh_command))
                retry -= 1
                if retry < 0:
                    LOG.error('It is time out, When execute command: '
                              '{}.'.format(ssh_command))
                    raise exceptions.MgmtDriverOtherError(
                        error_message='It is time out, When execute command: '
                                      '{}.'.format(ssh_command))
                time.sleep(COMMAND_WAIT_RETRY_TIME)
        if type == 'common':
            if result.get_return_code() != 0:
                err = result.get_stderr()
                LOG.error(err)
                raise exceptions.MgmtDriverRemoteCommandError(err_info=err)

        return result.get_stdout()

    def _get_nodeport_from_kubernetes(self, no_port_info_list, lb_commander,
                                      resource_info_list):
        for no_port_info in no_port_info_list:
            ssh_command = "kubectl describe svc '%(svc_name)s' -n" \
                          " '%(namespace)s' | grep NodePort: | awk" \
                          " '{print $3}' | awk -F '/' '{print $1}'" \
                          % {'svc_name': no_port_info.get('name'),
                             'namespace': no_port_info.get('namespace')}
            results = self._execute_command(
                lb_commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
            node_ports = ','.join([result.replace(
                '\n', '') for result in results])
            no_port_info['node_ports'] = node_ports
            resource_info_list.append(no_port_info)

    def _get_script_input_parameter(self, vnf_package_path, additional_param,
                                    operation_type):
        script_path = additional_param.get(
            'lcm-kubernetes-external-lb', {}).get('script_path')
        ssh_ip = additional_param.get(
            'lcm-kubernetes-external-lb', {}).get('external_lb_param').get(
            'ssh_ip')
        ssh_username = additional_param.get(
            'lcm-kubernetes-external-lb', {}).get('external_lb_param').get(
            'ssh_username')
        ssh_password = additional_param.get(
            'lcm-kubernetes-external-lb', {}).get('external_lb_param').get(
            'ssh_password')
        artifact_files = additional_param.get(
            'lcm-kubernetes-def-files', {})
        resource_info_list = []
        no_port_info_list = []
        for artifact_file in artifact_files:
            artiface_file_path = os.path.join(
                vnf_package_path, artifact_file)
            with open(artiface_file_path, 'r', encoding='utf-8') as f:
                yaml_content_all = yaml.safe_load_all(f.read())
            for yaml_content in yaml_content_all:
                if yaml_content.get('kind') == 'Service' and \
                        yaml_content.get('spec').get('type') == 'NodePort':
                    if operation_type == 'INSTANTIATE':
                        ports = yaml_content.get('spec').get('ports')
                        node_ports = [port.get(
                            'nodePort') for port in ports if port.get(
                            'nodePort')]
                        if len(node_ports) == len(ports):
                            node_ports_str = ','.join([str(
                                port) for port in node_ports])
                            resource_info = {
                                "namespace": yaml_content.get('metadata').get(
                                    'namespace', 'default'),
                                "name": yaml_content.get('metadata').get(
                                    'name'),
                                "node_ports": node_ports_str
                            }
                            resource_info_list.append(resource_info)
                        else:
                            no_port_info = {
                                "namespace": yaml_content.get('metadata').get(
                                    'namespace', 'default'),
                                "name": yaml_content.get('metadata').get(
                                    'name'),
                            }
                            no_port_info_list.append(no_port_info)
                    else:
                        resource_info = {
                            "namespace": yaml_content.get('metadata').get(
                                'namespace', 'default'),
                            "name": yaml_content.get('metadata').get('name')
                        }
                        resource_info_list.append(resource_info)

        lb_commander = self._init_commander_and_set_script(
            ssh_username, ssh_password, ssh_ip, K8S_CMD_TIMEOUT,
            vnf_package_path=vnf_package_path,
            script_path=script_path)
        if operation_type == 'INSTANTIATE':
            # get nodeport info from kubernetes
            self._get_nodeport_from_kubernetes(
                no_port_info_list, lb_commander, resource_info_list)
        resource_info_str_list = []
        for resource_info in resource_info_list:
            resource_info_str = ','.join(
                [value for key, value in resource_info.items()])
            resource_info_str_list.append(resource_info_str)
        all_resource_info_str = '#'.join(resource_info_str_list)
        return lb_commander, all_resource_info_str

    @log.log
    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        additional_param = instantiate_vnf_request.additional_params
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)
        self._check_input_parameters(additional_param.get(
            'lcm-kubernetes-external-lb', {}), vnf_package_path)
        lb_commander, all_resource_info_str = \
            self._get_script_input_parameter(
                vnf_package_path, additional_param, 'INSTANTIATE')
        ssh_command = 'bash /tmp/configure_lb.sh -i {} -a True'.format(
            all_resource_info_str)
        self._execute_command(
            lb_commander, ssh_command, K8S_DEPLOY_TIMEOUT, 'common', 0)
        lb_commander.close_session()

    @log.log
    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    @log.log
    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)
        add_param = {}
        if hasattr(terminate_vnf_request, 'additional_params') and \
                terminate_vnf_request.additional_params:
            additional_params = terminate_vnf_request.additional_params
            lb_params_default = \
                vnf_instance.instantiated_vnf_info.additional_params.get(
                    'lcm-kubernetes-external-lb')
            add_param['lcm-kubernetes-external-lb'] = additional_params.get(
                'lcm-kubernetes-external-lb', lb_params_default)
            add_param['lcm-kubernetes-def-files'] = \
                vnf_instance.instantiated_vnf_info.additional_params.get(
                    'lcm-kubernetes-def-files')
        else:
            add_param = \
                vnf_instance.instantiated_vnf_info.additional_params

        lb_commander, all_resource_info_str = \
            self._get_script_input_parameter(
                vnf_package_path, add_param, 'TERMINATE')
        ssh_command = 'bash /tmp/configure_lb.sh -i {} -a False'.format(
            all_resource_info_str)
        self._execute_command(
            lb_commander, ssh_command, K8S_DEPLOY_TIMEOUT, 'common', 0)
        lb_commander.close_session()

    @log.log
    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        pass

    @log.log
    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        pass

    @log.log
    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        if not heal_vnf_request.vnfc_instance_id:
            self.terminate_end(
                context, vnf_instance, heal_vnf_request,
                grant, grant_request)
        else:
            pass

    @log.log
    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        if not heal_vnf_request.vnfc_instance_id:
            if hasattr(heal_vnf_request, 'additional_params') and \
                    heal_vnf_request.additional_params:
                lb_params_default = \
                    vnf_instance.instantiated_vnf_info.additional_params.get(
                        'lcm-kubernetes-external-lb')
                if not heal_vnf_request.additional_params.get(
                        'lcm-kubernetes-external-lb'):
                    heal_vnf_request.additional_params[
                        'lcm-kubernetes-external-lb'] = lb_params_default
                heal_vnf_request.additional_params[
                    'lcm-kubernetes-def-files'] = \
                    vnf_instance.instantiated_vnf_info.additional_params.get(
                        'lcm-kubernetes-def-files')
            else:
                heal_vnf_request.additional_params = \
                    vnf_instance.instantiated_vnf_info.additional_params
            self.instantiate_end(context, vnf_instance, heal_vnf_request,
                                 grant, grant_request)
        else:
            pass

    @log.log
    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    @log.log
    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""empty message

Revision ID: 4c31092895b8
Revises: 81ffa86020d
Create Date: 2014-08-01 11:48:10.319498

"""

# revision identifiers, used by Alembic.
revision = '4c31092895b8'
down_revision = '81ffa86020d'


def upgrade(active_plugins=None, options=None):
    pass



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker import objects
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


def _get_vim_connection_info(vnf_instance):
    vim_connection_infos = []

    for vim_conn_info in vnf_instance.vim_connection_info:
        new_vim_conn_info = vim_conn_info.obj_clone()
        vim_connection_infos.append(new_vim_conn_info)

    return vim_connection_infos


def _get_ext_managed_virtual_links(vnf_instantiated_info):
    ext_managed_virtual_links = []

    for ext_mng_vl_info in \
            vnf_instantiated_info.ext_managed_virtual_link_info:
        network_resource = ext_mng_vl_info.network_resource

        ext_mng_vl_data = objects.ExtManagedVirtualLinkData(
            id=ext_mng_vl_info.id,
            vnf_virtual_link_desc_id=ext_mng_vl_info.vnf_virtual_link_desc_id,
            resource_id=network_resource.resource_id,
            vim_connection_id=network_resource.vim_connection_id)

        ext_managed_virtual_links.append(ext_mng_vl_data)

    return ext_managed_virtual_links


def _get_cp_instance_id(ext_virtual_link_info_id, vnf_instantiated_info):
    cp_instances = []

    for vnf_vl_res_info in \
            vnf_instantiated_info.vnf_virtual_link_resource_info:
        if ext_virtual_link_info_id == \
                vnf_vl_res_info.vnf_virtual_link_desc_id:
            for vnf_link_port in vnf_vl_res_info.vnf_link_ports:
                cp_instances.append(vnf_link_port.cp_instance_id)

    return cp_instances


def _get_cp_data_from_vnfc_resource_info(cp_instance_list,
        vnf_instantiated_info):
    vnfc_cp_infos = []

    for vnfc_resource_info in vnf_instantiated_info.vnfc_resource_info:
        for vnfc_cp_info in vnfc_resource_info.vnfc_cp_info:
            if vnfc_cp_info.id in cp_instance_list:
                vnfc_cp_infos.append(vnfc_cp_info)

    return vnfc_cp_infos


def _get_link_ports(vnfc_cp_info_list, vnf_instantiated_info):
    ext_link_ports = []

    for vnfc_cp_info in vnfc_cp_info_list:
        for ext_vl_info in vnf_instantiated_info.ext_virtual_link_info:
            for ext_vl_port in ext_vl_info.ext_link_ports:
                if ext_vl_port.id != vnfc_cp_info.vnf_ext_cp_id:
                    continue

                resource_handle = ext_vl_port.resource_handle.obj_clone()

                ext_link_port_data = objects.ExtLinkPortData(
                    id=ext_vl_port.id,
                    resource_handle=resource_handle)
                ext_link_ports.append(ext_link_port_data)
                break

    return ext_link_ports


def _get_cp_protocol_data_list(ext_cp_info):
    cp_protocol_data_list = []

    def _get_ip_addresses(ip_addresses):
        ip_addresses_list = []
        for ip_address in ip_addresses:
            ip_address_data = IpAddressReq(
                type=ip_address.type,
                subnet_id=ip_address.subnet_id,
                fixed_addresses=ip_address.addresses,
                num_dynamic_addresses=1 if ip_address.is_dynamic else 0)

            ip_addresses_list.append(ip_address_data)

        return ip_addresses_list

    for cp_protocol_info in ext_cp_info.cp_protocol_info:
        if cp_protocol_info.ip_over_ethernet:
            ip_addresses = _get_ip_addresses(cp_protocol_info.
                ip_over_ethernet.ip_addresses)

            ip_over_ethernet_data = objects.IpOverEthernetAddressData(
                mac_address=cp_protocol_info.ip_over_ethernet.mac_address,
                ip_addresses=ip_addresses)
        else:
            ip_over_ethernet_data = None

        cp_protocol_data = objects.CpProtocolData(
            layer_protocol=cp_protocol_info.layer_protocol,
            ip_over_ethernet=ip_over_ethernet_data)

        cp_protocol_data_list.append(cp_protocol_data)

    return cp_protocol_data_list


def _get_cp_ids(vnfc_cp_info_list, vnf_instantiated_info):

    ext_cps = []
    for vnfc_cp_info in vnfc_cp_info_list:
        for ext_cp_info in vnf_instantiated_info.ext_cp_info:
            if ext_cp_info.cpd_id != vnfc_cp_info.cpd_id:
                continue

            ext_cp_configs = []
            vnf_ext_cp_data = objects.VnfExtCpData()
            vnf_ext_cp_data.cpd_id = ext_cp_info.cpd_id

            cp_protocol_data_list = _get_cp_protocol_data_list(ext_cp_info)
            # TODO(nitin-uikey) set cp_instance_id
            # and prepare 1-N objects of VnfExtCpConfig
            vnf_ext_cp_config = objects.VnfExtCpConfig(
                link_port_id=vnfc_cp_info.vnf_ext_cp_id,
                cp_protocol_data=cp_protocol_data_list)

            ext_cp_configs.append(vnf_ext_cp_config)

            vnf_ext_cp_data = objects.VnfExtCpData(
                cpd_id=ext_cp_info.cpd_id,
                cp_config=ext_cp_configs)
            ext_cps.append(vnf_ext_cp_data)
            break

    return ext_cps


def _get_ext_virtual_link_data(vnf_instantiated_info):

    ext_virtual_links = []

    for ext_vl_info in \
            vnf_instantiated_info.ext_virtual_link_info:
        resource_handle = ext_vl_info.resource_handle
        ext_vl_data = objects.ExtVirtualLinkData(
            id=ext_vl_info.id,
            resource_id=resource_handle.resource_id,
            vim_connection_id=resource_handle.vim_connection_id)

        # call vnf virtual link resource info
        cp_instances = _get_cp_instance_id(ext_vl_info.id,
            vnf_instantiated_info)

        # get cp info from vnfcresources
        vnfc_cp_infos = _get_cp_data_from_vnfc_resource_info(cp_instances,
            vnf_instantiated_info)

        ext_vl_data.ext_link_ports = _get_link_ports(vnfc_cp_infos,
            vnf_instantiated_info)

        # assign the data to extcp info and link port
        ext_vl_data.ext_cps = _get_cp_ids(vnfc_cp_infos,
            vnf_instantiated_info)

        ext_virtual_links.append(ext_vl_data)

    return ext_virtual_links


@base.TackerObjectRegistry.register
class InstantiateVnfRequest(base.TackerObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'flavour_id': fields.StringField(nullable=False),
        'instantiation_level_id': fields.StringField(nullable=True,
                                                     default=None),
        'ext_managed_virtual_links': fields.ListOfObjectsField(
            'ExtManagedVirtualLinkData', nullable=True, default=[]),
        'vim_connection_info': fields.ListOfObjectsField(
            'VimConnectionInfo', nullable=True, default=[]),
        'ext_virtual_links': fields.ListOfObjectsField(
            'ExtVirtualLinkData', nullable=True, default=[]),
        'additional_params': fields.DictOfNullableField(nullable=True,
            default={})
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_instantiate_vnf_req = super(
                InstantiateVnfRequest, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ext_managed_virtual_links' in primitive.keys():
                obj_data = [ExtManagedVirtualLinkData._from_dict(
                    ext_manage) for ext_manage in primitive.get(
                    'ext_managed_virtual_links', [])]
                primitive.update({'ext_managed_virtual_links': obj_data})

            if 'vim_connection_info' in primitive.keys():
                obj_data = [objects.VimConnectionInfo._from_dict(
                    vim_conn) for vim_conn in primitive.get(
                    'vim_connection_info', [])]
                primitive.update({'vim_connection_info': obj_data})

            if 'ext_virtual_links' in primitive.keys():
                obj_data = [ExtVirtualLinkData.obj_from_primitive(
                    ext_vir_link, context) for ext_vir_link in primitive.get(
                    'ext_virtual_links', [])]
                primitive.update({'ext_virtual_links': obj_data})
            obj_instantiate_vnf_req = InstantiateVnfRequest._from_dict(
                primitive)

        return obj_instantiate_vnf_req

    @classmethod
    def _from_dict(cls, data_dict):
        flavour_id = data_dict.get('flavour_id')
        instantiation_level_id = data_dict.get('instantiation_level_id')
        ext_managed_virtual_links = data_dict.get('ext_managed_virtual_links',
                                                  [])
        vim_connection_info = data_dict.get('vim_connection_info', [])
        ext_virtual_links = data_dict.get('ext_virtual_links', [])
        additional_params = data_dict.get('additional_params', {})

        return cls(flavour_id=flavour_id,
        instantiation_level_id=instantiation_level_id,
        ext_managed_virtual_links=ext_managed_virtual_links,
        vim_connection_info=vim_connection_info,
        ext_virtual_links=ext_virtual_links,
        additional_params=additional_params)

    @classmethod
    def from_vnf_instance(cls, vnf_instance):

        vnf_instantiated_info = vnf_instance.instantiated_vnf_info

        # Vim connection info
        vim_connection_info = _get_vim_connection_info(vnf_instance)

        # Flavour id
        flavour_id = vnf_instantiated_info.flavour_id

        # Instantiation level
        instantiation_level_id = vnf_instantiated_info.instantiation_level_id

        # Externally managed virtual links
        ext_managed_virtual_links = _get_ext_managed_virtual_links(
            vnf_instantiated_info)

        # External virtual links
        ext_virtual_links = _get_ext_virtual_link_data(vnf_instantiated_info)

        additional_params = vnf_instantiated_info.additional_params

        instantiate_vnf_request = cls(flavour_id=flavour_id,
            instantiation_level_id=instantiation_level_id,
            ext_managed_virtual_links=ext_managed_virtual_links,
            vim_connection_info=vim_connection_info,
            ext_virtual_links=ext_virtual_links,
            additional_params=additional_params)

        return instantiate_vnf_request


@base.TackerObjectRegistry.register
class ExtManagedVirtualLinkData(base.TackerObject):
    # Version 1.0: Initial version
    # Version 1.1: Added field for vim_connection_id
    VERSION = '1.1'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnf_virtual_link_desc_id': fields.StringField(nullable=False),
        'resource_id': fields.StringField(nullable=False),
        'vim_connection_id': fields.StringField(nullable=True)
    }

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vnf_virtual_link_desc_id = data_dict.get(
            'vnf_virtual_link_desc_id')
        resource_id = data_dict.get('resource_id')
        vim_connection_id = data_dict.get('vim_connection_id')
        obj = cls(id=id, vnf_virtual_link_desc_id=vnf_virtual_link_desc_id,
                  resource_id=resource_id, vim_connection_id=vim_connection_id)
        return obj


@base.TackerObjectRegistry.register
class ExtVirtualLinkData(base.TackerObject):
    # Version 1.0: Initial version
    # Version 1.1: Added field for vim_connection_id
    VERSION = '1.1'

    fields = {
        'id': fields.StringField(nullable=False),
        'resource_id': fields.StringField(nullable=False),
        'vim_connection_id': fields.StringField(nullable=True),
        'ext_cps': fields.ListOfObjectsField(
            'VnfExtCpData', nullable=True, default=[]),
        'ext_link_ports': fields.ListOfObjectsField(
            'ExtLinkPortData', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_ext_virt_link = super(
                ExtVirtualLinkData, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ext_cps' in primitive.keys():
                obj_data = [VnfExtCpData.obj_from_primitive(
                    ext_cp, context) for ext_cp in primitive.get(
                    'ext_cps', [])]
                primitive.update({'ext_cps': obj_data})

            if 'ext_link_ports' in primitive.keys():
                obj_data = [ExtLinkPortData.obj_from_primitive(
                    ext_link_port_data, context)
                    for ext_link_port_data in primitive.get(
                        'ext_link_ports', [])]
                primitive.update({'ext_link_ports': obj_data})

            obj_ext_virt_link = ExtVirtualLinkData._from_dict(primitive)

        return obj_ext_virt_link

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        resource_id = data_dict.get('resource_id')
        vim_connection_id = data_dict.get('vim_connection_id')
        ext_cps = data_dict.get('ext_cps', [])
        ext_link_ports = data_dict.get('ext_link_ports', [])

        obj = cls(id=id, resource_id=resource_id, ext_cps=ext_cps,
                  ext_link_ports=ext_link_ports,
                  vim_connection_id=vim_connection_id)
        return obj


@base.TackerObjectRegistry.register
class VnfExtCpData(base.TackerObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'cpd_id': fields.StringField(nullable=False),
        'cp_config': fields.ListOfObjectsField(
            'VnfExtCpConfig', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_vnf_ext_cp_data = super(VnfExtCpData, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'cp_config' in primitive.keys():
                obj_data = [VnfExtCpConfig.obj_from_primitive(
                    vnf_ext_cp_conf, context)
                    for vnf_ext_cp_conf in primitive.get('cp_config', [])]
                primitive.update({'cp_config': obj_data})

            obj_vnf_ext_cp_data = VnfExtCpData._from_dict(primitive)

        return obj_vnf_ext_cp_data

    @classmethod
    def _from_dict(cls, data_dict):
        cpd_id = data_dict.get('cpd_id')
        cp_config = data_dict.get('cp_config', [])

        obj = cls(cpd_id=cpd_id, cp_config=cp_config)
        return obj


@base.TackerObjectRegistry.register
class VnfExtCpConfig(base.TackerObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'cp_instance_id': fields.StringField(nullable=True, default=None),
        'link_port_id': fields.StringField(nullable=True, default=None),
        'cp_protocol_data': fields.ListOfObjectsField(
            'CpProtocolData', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_ext_cp_config = super(VnfExtCpConfig, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'cp_protocol_data' in primitive.keys():
                obj_data = [CpProtocolData.obj_from_primitive(
                    cp_protocol, context) for cp_protocol in primitive.get(
                    'cp_protocol_data', [])]
                primitive.update({'cp_protocol_data': obj_data})

            obj_ext_cp_config = VnfExtCpConfig._from_dict(primitive)

        return obj_ext_cp_config

    @classmethod
    def _from_dict(cls, data_dict):
        cp_instance_id = data_dict.get('cp_instance_id')
        link_port_id = data_dict.get('link_port_id')
        cp_protocol_data = data_dict.get('cp_protocol_data', [])

        obj = cls(cp_instance_id=cp_instance_id,
                  link_port_id=link_port_id, cp_protocol_data=cp_protocol_data)
        return obj


@base.TackerObjectRegistry.register
class CpProtocolData(base.TackerObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'layer_protocol': fields.StringField(nullable=False),
        'ip_over_ethernet': fields.ObjectField(
            'IpOverEthernetAddressData', nullable=True, default=None),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_cp_protocal = super(CpProtocolData, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ip_over_ethernet' in primitive.keys():
                obj_data = IpOverEthernetAddressData.obj_from_primitive(
                    primitive.get('ip_over_ethernet', {}), context)
                primitive.update({'ip_over_ethernet': obj_data})
            obj_cp_protocal = CpProtocolData._from_dict(primitive)

        return obj_cp_protocal

    @classmethod
    def _from_dict(cls, data_dict):
        layer_protocol = data_dict.get('layer_protocol')
        ip_over_ethernet = data_dict.get('ip_over_ethernet')

        obj = cls(layer_protocol=layer_protocol,
                  ip_over_ethernet=ip_over_ethernet)
        return obj


@base.TackerObjectRegistry.register
class IpOverEthernetAddressData(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'mac_address': fields.StringField(
            nullable=True,
            default=None),
        'ip_addresses': fields.ListOfObjectsField(
            'IpAddressReq',
            nullable=True,
            default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            ip_over_ethernet = super(
                IpOverEthernetAddressData, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ip_addresses' in primitive.keys():
                obj_data = [IpAddressReq._from_dict(
                    ip_address) for ip_address in primitive.get(
                        'ip_addresses', [])]
                primitive.update({'ip_addresses': obj_data})

            ip_over_ethernet = IpOverEthernetAddressData._from_dict(primitive)

        return ip_over_ethernet

    @classmethod
    def _from_dict(cls, data_dict):
        mac_address = data_dict.get('mac_address')
        ip_addresses = data_dict.get('ip_addresses', [])
        obj = cls(mac_address=mac_address, ip_addresses=ip_addresses)
        return obj


@base.TackerObjectRegistry.register
class IpAddressReq(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'type': fields.IpAddressTypeField(nullable=False),
        'subnet_id': fields.StringField(nullable=True, default=None),
        'fixed_addresses': fields.ListOfStringsField(nullable=True,
            default=[]),
        'num_dynamic_addresses': fields.IntegerField(nullable=True, default=0)
    }

    @classmethod
    def _from_dict(cls, data_dict):
        type = data_dict.get('type')
        subnet_id = data_dict.get('subnet_id')
        fixed_addresses = data_dict.get('fixed_addresses', [])
        num_dynamic_addresses = data_dict.get('num_dynamic_addresses')

        obj = cls(type=type, subnet_id=subnet_id,
                  fixed_addresses=fixed_addresses,
                  num_dynamic_addresses=num_dynamic_addresses)

        return obj


@base.TackerObjectRegistry.register
class ExtLinkPortData(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'resource_handle': fields.ObjectField(
            'ResourceHandle', nullable=False),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_link_port_data = super(
                ExtLinkPortData, cls).obj_from_primitive(primitive, context)
        else:
            if 'resource_handle' in primitive.keys():
                obj_data = objects.ResourceHandle._from_dict(primitive.get(
                    'resource_handle', []))
                primitive.update({'resource_handle': obj_data})

            obj_link_port_data = ExtLinkPortData._from_dict(primitive)

        return obj_link_port_data

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        resource_handle = data_dict.get('resource_handle')

        obj = cls(id=id, resource_handle=resource_handle)
        return obj



#    Copyright 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.common import exceptions
import tacker.context
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _vnf_package_vnfd_create(context, values):
    vnf_package_vnfd = models.VnfPackageVnfd()

    vnf_package_vnfd.update(values)
    try:
        vnf_package_vnfd.save(context.session)
    except db_exc.DBDuplicateEntry as e:
        if 'vnfd_id' in e.columns:
            raise exceptions.VnfPackageVnfdIdDuplicate(
                vnfd_id=values.get('vnfd_id'))

    return vnf_package_vnfd


@db_api.context_manager.reader
def _get_vnf_package_vnfd(context, id, package_uuid=None, del_flg=None):
    if package_uuid and not del_flg:
        query = api.model_query(
            context,
            models.VnfPackageVnfd).filter_by(
            package_uuid=id).filter_by(
            deleted=0)
    elif package_uuid and del_flg:
        query = api.model_query(
            context, models.VnfPackageVnfd).filter_by(
            package_uuid=id)
    else:
        query = api.model_query(
            context,
            models.VnfPackageVnfd).filter_by(
            vnfd_id=id).filter_by(
            deleted=0)
    try:
        result = query.all()
        result_line = ""
        for line in result:
            result_line = line

    except Exception:
        LOG.error("select vnf_package_vnfd failed")

    if result_line:
        return result_line
    else:
        return None


@db_api.context_manager.writer
def _vnf_package_vnfd_delete(context, id):
    try:
        api.model_query(
            context, models.VnfPackageVnfd).filter_by(
            package_uuid=id).delete()
    except Exception:
        LOG.error("delete vnf_package_vnfd failed")


@db_api.context_manager.reader
def _vnf_package_vnfd_get_by_id(context, vnfd_id):

    query = api.model_query(context, models.VnfPackageVnfd,
                            read_deleted="no", project_only=False). \
        filter_by(vnfd_id=vnfd_id).\
        join((models.VnfPackage, models.VnfPackage.id ==
              models.VnfPackageVnfd.package_uuid))

    if tacker.context.is_user_context(context):
        query = query.filter(models.VnfPackage.tenant_id == context.project_id)

    result = query.first()

    if not result:
        raise exceptions.VnfPackageVnfdNotFound(id=vnfd_id)

    return result


def _vnf_package_vnfd_get_by_packageId(context, packageId):

    query = api.model_query(
        context,
        models.VnfPackageVnfd,
        read_deleted="no",
        project_only=True).filter_by(
        package_uuid=packageId)

    result = query.first()

    if not result:
        return None

    return result


@db_api.context_manager.reader
def _vnf_package_vnfd_get_by_vnfdId(context, vnfdId):
    query = api.model_query(context,
                            models.VnfPackageVnfd,
                            read_deleted="no",
                            project_only=True).filter_by(vnfd_id=vnfdId)

    result = query.first()

    if not result:
        return None

    return result


@db_api.context_manager.reader
def _get_vnf_package_vnfd_by_vnfid(context, vnfpkgid):

    sql = ("select"
           " t1.vnfd_id,"
           " t1.vnf_provider,"
           " t1.vnf_product_name,"
           " t1.vnf_software_version,"
           " t1.vnfd_version,"
           " t2.name"
           " from "
           " vnf_package_vnfd t1,"
           " vnf t2 "
           " where"
           " t1.vnfd_id=t2.vnfd_id"
           " and"
           " t2.id= :vnfpkgid")

    result = context.session.execute(sql, {'vnfpkgid': vnfpkgid})
    for line in result:
        return line


@base.TackerObjectRegistry.register
class VnfPackageVnfd(base.TackerObject, base.TackerObjectDictCompat,
                     base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'package_uuid': fields.UUIDField(nullable=False),
        'vnfd_id': fields.UUIDField(nullable=False),
        'vnf_provider': fields.StringField(nullable=False),
        'vnf_product_name': fields.StringField(nullable=False),
        'vnf_software_version': fields.StringField(nullable=False),
        'vnfd_version': fields.StringField(nullable=False),
    }

    @staticmethod
    def _from_db_object(context, vnf_package_vnfd, db_vnf_package_vnfd):

        for key in vnf_package_vnfd.fields:
            if db_vnf_package_vnfd[key]:
                setattr(vnf_package_vnfd, key, db_vnf_package_vnfd[key])

        vnf_package_vnfd._context = context
        vnf_package_vnfd.obj_reset_changes()

        return vnf_package_vnfd

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='create',
                                               reason=_('already created'))
        updates = self.obj_get_changes()

        if 'id' not in updates:
            updates['id'] = uuidutils.generate_uuid()
            self.id = updates['id']

        updates = self.obj_get_changes()
        db_vnf_package_vnfd = _vnf_package_vnfd_create(
            self._context, updates)
        self._from_db_object(self._context, self, db_vnf_package_vnfd)

    @base.remotable
    def get_vnf_package_vnfd(self, id, package_uuid=None, del_flg=None):
        return _get_vnf_package_vnfd(self._context, id, package_uuid, del_flg)

    @base.remotable_classmethod
    def get_vnf_package_vnfd_by_vnfid(self, context, vnfid):
        return _get_vnf_package_vnfd_by_vnfid(context, vnfid)

    @base.remotable
    def delete(self, id):
        _vnf_package_vnfd_delete(self._context, id)

    @classmethod
    def obj_from_db_obj(cls, context, db_obj):
        return cls._from_db_object(context, cls(), db_obj)

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_vnf_package_vnfd = _vnf_package_vnfd_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_vnf_package_vnfd)

    @base.remotable_classmethod
    def get_by_vnfdId(cls, context, id):
        db_vnf_package_vnfd = _vnf_package_vnfd_get_by_vnfdId(
            context, id)
        if not db_vnf_package_vnfd:
            return db_vnf_package_vnfd
        return cls._from_db_object(context, cls(), db_vnf_package_vnfd)

    @base.remotable_classmethod
    def get_by_packageId(cls, context, id):
        db_vnf_package_vnfd = _vnf_package_vnfd_get_by_packageId(
            context, id)
        if not db_vnf_package_vnfd:
            return db_vnf_package_vnfd
        return cls._from_db_object(context, cls(), db_vnf_package_vnfd)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""remove infra_driver column

Revision ID: 8f7145914cb0
Revises: 0ae5b1ce3024
Create Date: 2016-12-08 17:28:26.609343

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '8f7145914cb0'
down_revision = '0ae5b1ce3024'

from alembic import op


def upgrade(active_plugins=None, options=None):
    op.drop_column('vnfd', 'infra_driver')



#    Copyright 2018 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.objects import base
from tacker.objects import fields
from tacker.tests.unit import base as test_base


class TestString(test_base.TestCase):
    def setUp(self):
        super(TestString, self).setUp()
        self.field = fields.StringField()
        self.coerce_good_values = [('foo', 'foo'), (1, '1'), (True, 'True')]
        self.coerce_bad_values = [None]

    def test_stringify(self):
        self.assertEqual("'123'", self.field.stringify(123))


class TestListOfStrings(test_base.TestCase):
    def setUp(self):
        super(TestListOfStrings, self).setUp()
        self.field = fields.ListOfStringsField()

    def test_list_of_string(self):
        self.assertEqual("['abc']", self.field.stringify(['abc']))


class TestListOfObjects(test_base.TestCase):

    def test_list_of_obj(self):
        @base.TackerObjectRegistry.register_if(False)
        class MyObjElement(base.TackerObject):
            fields = {'foo': fields.StringField()}

            def __init__(self, foo):
                super(MyObjElement, self).__init__()
                self.foo = foo

        @base.TackerObjectRegistry.register_if(False)
        class MyList(base.TackerObject):
            fields = {'objects': fields.ListOfObjectsField('MyObjElement')}

        mylist = MyList()
        mylist.objects = [MyObjElement('a'), MyObjElement('b')]
        self.assertEqual(['a', 'b'], [x.foo for x in mylist.objects])



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""adds_VNFFG

Revision ID: 507122918800
Revises: 4ee19c8a6d0a
Create Date: 2016-07-29 21:48:18.816277

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '507122918800'
down_revision = '4ee19c8a6d0a'

import sqlalchemy as sa

from alembic import op
from tacker.db.types import Json


def upgrade(active_plugins=None, options=None):

    op.create_table(
        'vnffgtemplates',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('template', Json),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'vnffgs',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('vnffgd_id', sa.String(length=36), nullable=False),
        sa.Column('status', sa.String(length=255), nullable=False),
        sa.Column('vnf_mapping', Json),
        sa.ForeignKeyConstraint(['vnffgd_id'], ['vnffgtemplates.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'vnffgnfps',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('vnffg_id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('status', sa.String(length=255), nullable=False),
        sa.Column('path_id', sa.String(length=255), nullable=False),
        sa.Column('symmetrical', sa.Boolean, default=False),
        sa.ForeignKeyConstraint(['vnffg_id'], ['vnffgs.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'vnffgchains',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('instance_id', sa.String(length=255), nullable=True),
        sa.Column('nfp_id', sa.String(length=36), nullable=False),
        sa.Column('status', sa.String(length=255), nullable=False),
        sa.Column('path_id', sa.String(length=255), nullable=False),
        sa.Column('symmetrical', sa.Boolean, default=False),
        sa.Column('chain', Json),
        sa.ForeignKeyConstraint(['nfp_id'], ['vnffgnfps.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'vnffgclassifiers',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('nfp_id', sa.String(length=36), nullable=False),
        sa.Column('instance_id', sa.String(length=255), nullable=True),
        sa.Column('chain_id', sa.String(length=36), nullable=False),
        sa.Column('status', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['nfp_id'], ['vnffgnfps.id'], ),
        sa.ForeignKeyConstraint(['chain_id'], ['vnffgchains.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'aclmatchcriterias',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vnffgc_id', sa.String(length=36), nullable=False),
        sa.Column('eth_src', sa.String(length=36), nullable=True),
        sa.Column('eth_dst', sa.String(length=36), nullable=True),
        sa.Column('eth_type', sa.String(length=36), nullable=True),
        sa.Column('vlan_id', sa.Integer, nullable=True),
        sa.Column('vlan_pcp', sa.Integer, nullable=True),
        sa.Column('mpls_label', sa.Integer, nullable=True),
        sa.Column('mpls_tc', sa.Integer, nullable=True),
        sa.Column('ip_dscp', sa.Integer, nullable=True),
        sa.Column('ip_ecn', sa.Integer, nullable=True),
        sa.Column('ip_src_prefix', sa.String(length=36), nullable=True),
        sa.Column('ip_dst_prefix', sa.String(length=36), nullable=True),
        sa.Column('source_port_min', sa.Integer, nullable=True),
        sa.Column('source_port_max', sa.Integer, nullable=True),
        sa.Column('destination_port_min', sa.Integer, nullable=True),
        sa.Column('destination_port_max', sa.Integer, nullable=True),
        sa.Column('ip_proto', sa.Integer, nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('network_src_port_id', sa.String(length=36), nullable=True),
        sa.Column('network_dst_port_id', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=64), nullable=True),
        sa.Column('icmpv4_type', sa.Integer, nullable=True),
        sa.Column('icmpv4_code', sa.Integer, nullable=True),
        sa.Column('arp_op', sa.Integer, nullable=True),
        sa.Column('arp_spa', sa.Integer, nullable=True),
        sa.Column('arp_tpa', sa.Integer, nullable=True),
        sa.Column('arp_sha', sa.Integer, nullable=True),
        sa.Column('arp_tha', sa.Integer, nullable=True),
        sa.Column('ipv6_src', sa.String(36), nullable=True),
        sa.Column('ipv6_dst', sa.String(36), nullable=True),
        sa.Column('ipv6_flabel', sa.Integer, nullable=True),
        sa.Column('icmpv6_type', sa.Integer, nullable=True),
        sa.Column('icmpv6_code', sa.Integer, nullable=True),
        sa.Column('ipv6_nd_target', sa.String(36), nullable=True),
        sa.Column('ipv6_nd_sll', sa.String(36), nullable=True),
        sa.Column('ipv6_nd_tll', sa.String(36), nullable=True),
        sa.ForeignKeyConstraint(['vnffgc_id'], ['vnffgclassifiers.id'], ),
        sa.PrimaryKeyConstraint('id'),
    )



# Copyright 2016 Brocade Communications Systems Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker._i18n import _
from tacker.api import extensions
from tacker.api.v1 import attributes as attr
from tacker.api.v1 import resource_helper
from tacker.common import exceptions
from tacker.plugins.common import constants
from tacker.services import service_base


class EventCreationFailureException(exceptions.TackerException):
    message = _("Failed to create an event: %(error_str)s")


class EventNotFoundException(exceptions.TackerException):
    message = _("Specified Event id %(evt_id)s is invalid. Please verify and "
                "pass a valid Event id")


class InvalidModelException(exceptions.TackerException):
    message = _("Specified model is invalid, only Event model supported")


class InputValuesMissing(exceptions.InvalidInput):
    message = _("Parameter input values missing for the key '%(key)s'")


class ParamYAMLInputMissing(exceptions.InvalidInput):
    message = _("Parameter YAML input missing")


class InvalidFormat(exceptions.InvalidInput):
    message = _("Invalid format. '%(error)s'")


RESOURCE_ATTRIBUTE_MAP = {

    'events': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'resource_id': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True
        },
        'resource_type': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True
        },
        'resource_state': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True
        },
        'timestamp': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'event_details': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'event_type': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
    }
}


class Common_services(extensions.ExtensionDescriptor):
    @classmethod
    def get_name(cls):
        return 'COMMONSERVICES'

    @classmethod
    def get_alias(cls):
        return 'Commonservices'

    @classmethod
    def get_description(cls):
        return "Extension for CommonServices"

    @classmethod
    def get_namespace(cls):
        return 'https://wiki.openstack.org/Tacker'

    @classmethod
    def get_updated(cls):
        return "2016-06-06T13:00:00-00:00"

    @classmethod
    def get_resources(cls):
        special_mappings = {}
        plural_mappings = resource_helper.build_plural_mappings(
            special_mappings, RESOURCE_ATTRIBUTE_MAP)
        attr.PLURALS.update(plural_mappings)
        return resource_helper.build_resource_info(
            plural_mappings, RESOURCE_ATTRIBUTE_MAP, constants.COMMONSERVICES,
            translate_name=True)

    @classmethod
    def get_plugin_interface(cls):
        return CommonServicesPluginBase

    def update_attributes_map(self, attributes):
        super(Common_services, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        version_map = {'1.0': RESOURCE_ATTRIBUTE_MAP}
        return version_map.get(version, {})


class CommonServicesPluginBase(service_base.NFVPluginBase,
        metaclass=abc.ABCMeta):
    def get_plugin_name(self):
        return constants.COMMONSERVICES

    def get_plugin_type(self):
        return constants.COMMONSERVICES

    def get_plugin_description(self):
        return 'Tacker CommonServices plugin'

    @abc.abstractmethod
    def get_event(self, context, event_id, fields=None):
        pass

    @abc.abstractmethod
    def get_events(self, context, filters=None, fields=None, sorts=None,
                   limit=None, marker_obj=None, page_reverse=False):
        pass



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


CONF = cfg.CONF

OPTS = [
    cfg.StrOpt('vnf_package_csar_path',
               default='/var/lib/tacker/vnfpackages/',
               help="Path to store extracted CSAR file"),

    cfg.FloatOpt('csar_file_size_cap', default=1024, min=0.000001,
               max=9223372036,
               help=_("""
Maximum size of CSAR file a user can upload in GB.

An CSAR file upload greater than the size mentioned here would result
in an CSAR upload failure. This configuration option defaults to
1024 GB (1 TiB).

NOTES:
    * This value should only be increased after careful
      consideration and must be set less than or equal to
      8 EiB (~9223372036).
    * This value must be set with careful consideration of the
      backend storage capacity. Setting this to a very low value
      may result in a large number of image failures. And, setting
      this to a very large value may result in faster consumption
      of storage. Hence, this must be set according to the nature of
      images created and storage capacity available.

Possible values:
    * Any positive number less than or equal to 9223372036854775808
""")),
    cfg.StrOpt('hashing_algorithm',
               default='sha512',
               help=_("""
Secure hashing algorithm used for computing the 'hash' property.

Possible values:
    * sha256, sha512

Related options:
    * None
""")),

    cfg.ListOpt('get_top_list',
                default=['tosca_definitions_version',
                    'description', 'metadata'],
                help=_("List of items to get from top-vnfd")),

    cfg.ListOpt('exclude_node',
                default=['VNF'],
                help=_("Exclude node from node_template")),

    cfg.ListOpt('get_lower_list',
                default=['tosca.nodes.nfv.VNF', 'tosca.nodes.nfv.VDU.Tacker'],
                help=_("List of types to get from lower-vnfd")),

    cfg.ListOpt('del_input_list',
                default=['descriptor_id', 'descriptor_version'
                'provider', 'product_name', 'software_version',
                'vnfm_info', 'flavour_id', 'flavour_description'],
                help=_("List of del inputs from lower-vnfd")),

]

vnf_package_group = cfg.OptGroup('vnf_package',
    title='vnf_package options',
    help="""
Options under this group are used to store vnf packages in glance store.
""")


def register_opts(conf):
    conf.register_group(vnf_package_group)
    conf.register_opts(OPTS, group=vnf_package_group)


def list_opts():
    return {vnf_package_group: OPTS}



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

CONF = cfg.CONF

VNFM_OPTS = [
    cfg.StrOpt('endpoint',
               default='http://127.0.0.1:9890',
               help='Endpoint of VNFM (self).'),
    cfg.IntOpt('default_graceful_termination_timeout',
               default=10,
               help='Default timeout value (second) of GRACEFUL termination.'),
    cfg.IntOpt('max_content_length',
               default=1000000,
               help='Max content length for list APIs.'),
    cfg.IntOpt('openstack_vim_stack_create_timeout',
               default=20,
               help='Timeout (in minuts) of heat stack creation.'),
    # NOTE: This is for test use since it is convenient to be able to delete
    # under development.
    cfg.BoolOpt('test_enable_lcm_op_occ_delete',
                default=False,
                help=_('Enable to delete LCM operation occurrence if True. '
                       'This is intended to use under development.')),
]

CONF.register_opts(VNFM_OPTS, 'v2_vnfm')

NFVO_OPTS = [
    cfg.BoolOpt('use_external_nfvo',
                default=False,
                help=_('Use external NFVO if True, '
                       'use internal NFVO in tacker if False')),
    cfg.StrOpt('grant_api_version',
               default='1.4.0',  # SOL003 v3.3.1 9.1a
               help='Grant api_version of NFVO.'),
    cfg.StrOpt('vnfpkgm_api_version',
               default='2.1.0',  # SOL003 v3.3.1 10.1a
               help='Vnf package management api_version of NFVO.'),
    # The following four parameters are for external NFVO.
    # Must be set when using external NFVO.
    # NOTE: It is assumed the VNFM communicates only one NFVO. That is
    # the same NFVO provides both the grant and vnf package management APIs.
    cfg.StrOpt('endpoint',
               default='',
               help='Endpoint of external NFVO.'),
    cfg.StrOpt('token_endpoint',
               default='',
               help='Token endpoint for OAuth2.0 authentication.'),
    cfg.StrOpt('client_id',
               default='',
               help='Client id used by OAuth2.0 authentication.'),
    cfg.StrOpt('client_password',
               default='',
               help='Client password used by OAuth2.0 authentication.'),
    cfg.BoolOpt('test_callback_uri',
                default=True,
                help='Check to get notification from callback Uri.'),
    cfg.ListOpt('test_grant_zone_list',
                default=["nova"],
                help='Zones used for test which returned in Grant response.')
]

CONF.register_opts(NFVO_OPTS, 'v2_nfvo')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils

from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfLcmOpOcc(SqlTestCase):

    def setUp(self):
        super(TestVnfLcmOpOcc, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_vnf_package()
        self.vnf_package_vnfd = self._create_and_upload_vnf_package_vnfd()
        self.vnf_instance = self._create_vnf_instance()
        self.vnf_lcm_op_occs = self._create_vnf_lcm_op_occs()

    def _create_vnf_package(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        return vnfpkgm

    def _create_and_upload_vnf_package_vnfd(self):
        vnf_package = objects.VnfPackage(context=self.context,
                                         **fakes.vnf_package_data)
        vnf_package.create()

        vnf_pack_vnfd = fakes.get_vnf_package_vnfd_data(
            vnf_package.id, uuidsentinel.vnfd_id)

        vnf_pack_vnfd_obj = objects.VnfPackageVnfd(
            context=self.context, **vnf_pack_vnfd)
        vnf_pack_vnfd_obj.create()

        return vnf_pack_vnfd_obj

    def _create_vnf_instance(self):
        vnf_instance_data = fakes.get_vnf_instance_data(
            self.vnf_package_vnfd.vnfd_id)
        vnf_instance = objects.VnfInstance(context=self.context,
                                           **vnf_instance_data)
        vnf_instance.create()

        return vnf_instance

    def _create_vnf_lcm_op_occs(self):
        id = uuidutils.generate_uuid()
        vnf_lcm_op_occs_data = \
            fakes.get_lcm_op_occs_data(id, self.vnf_instance.id)
        vnf_lcm_op_occs = objects.vnf_lcm_op_occs.VnfLcmOpOcc(
            context=self.context, **vnf_lcm_op_occs_data)
        vnf_lcm_op_occs.create()
        return vnf_lcm_op_occs

    def test_create(self):
        id = uuidutils.generate_uuid()
        vnf_lcm_op_occs_data = \
            fakes.get_lcm_op_occs_data(id, self.vnf_instance.id)
        vnf_lcm_op_occs = objects.vnf_lcm_op_occs.VnfLcmOpOcc(
            context=self.context, **vnf_lcm_op_occs_data)
        vnf_lcm_op_occs.create()
        self.assertTrue(vnf_lcm_op_occs.vnf_instance_id)
        self.assertTrue(vnf_lcm_op_occs.tenant_id)

    def test_save(self):
        id = uuidutils.generate_uuid()
        vnf_lcm_op_occs_data = \
            fakes.get_lcm_op_occs_data(id, self.vnf_instance.id)
        vnf_lcm_op_occs = objects.vnf_lcm_op_occs.VnfLcmOpOcc(
            context=self.context, **vnf_lcm_op_occs_data)
        vnf_lcm_op_occs.create()

        problem_obj = objects.vnf_lcm_op_occs.ProblemDetails()
        problem_obj.status = '500'
        problem_obj.detail = 'test_err'
        changed_info = objects.vnf_lcm_op_occs.VnfInfoModifications(
            context=self.context, **fakes.get_changed_info_data())
        changed_ext_conn = [objects.ExtVirtualLinkInfo.obj_from_primitive(
            i, self.context) for i in fakes.get_changed_ext_conn_data()]
        vnf_lcm_op_occs.operation_state = 'FAILED_TEMP'
        vnf_lcm_op_occs.error = problem_obj
        vnf_lcm_op_occs.id = uuidsentinel.vnf_lcm_op_occs_id
        vnf_lcm_op_occs.changed_info = changed_info
        vnf_lcm_op_occs.changed_ext_connectivity = changed_ext_conn
        vnf_lcm_op_occs.save()

        self.assertEqual('FAILED_TEMP', vnf_lcm_op_occs.operation_state)
        self.assertEqual(problem_obj, vnf_lcm_op_occs.error)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.api import views as base
from tacker.common import utils
import tacker.conf
from tacker.objects import vnf_lcm_op_occs as _vnf_lcm_op_occs

CONF = tacker.conf.CONF

LOG = logging.getLogger(__name__)


class ViewBuilder(base.BaseViewBuilder):

    FLATTEN_ATTRIBUTES = _vnf_lcm_op_occs.VnfLcmOpOcc.FLATTEN_ATTRIBUTES
    COMPLEX_ATTRIBUTES = _vnf_lcm_op_occs.VnfLcmOpOcc.COMPLEX_ATTRIBUTES
    FLATTEN_COMPLEX_ATTRIBUTES = [key for key in FLATTEN_ATTRIBUTES.keys()
        if '/' in key]

    def _get_lcm_op_occs_links(self, vnf_lcm_op_occs):
        _links = {
            "self": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_lcm_op_occs/{id}'.format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.id)
            },
            "vnfInstance": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_instances/{id}'.format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.vnf_instance_id)
            },
            "retry": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_lcm_op_occs/{id}/retry'.
                format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.id)
            },
            "rollback": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_lcm_op_occs/{id}/rollback'.
                format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.id)
            },
            "grant": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_lcm_op_occs/{id}/grant'.
                format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.id)
            },
            "fail": {
                "href":
                '{endpoint}/vnflcm/v1/vnf_lcm_op_occs/{id}/fail'.
                format(
                    endpoint=CONF.vnf_lcm.endpoint_url.rstrip("/"),
                    id=vnf_lcm_op_occs.id)
            }
        }

        return {"_links": _links}

    def _get_vnf_lcm_op_occs_list(self, vnf_lcm_op_occs, include_fields=None):
        vnf_lcm_op_occs_dict = vnf_lcm_op_occs.to_dict(
            include_fields=include_fields)

        vnf_lcm_op_occs_dict = utils.convert_snakecase_to_camelcase(
            vnf_lcm_op_occs_dict)
        vnf_lcm_op_occs_dict.pop('errorPoint', None)

        links = self._get_lcm_op_occs_links(vnf_lcm_op_occs)

        vnf_lcm_op_occs_dict.update(links)
        return vnf_lcm_op_occs_dict

    def index(self, request, vnf_lcm_op_occs, all_fields=True,
             exclude_fields=None, fields=None, exclude_default=False):

        # Find out which fields are to be returned in the response.
        if all_fields:
            include_fields = set(self.FLATTEN_ATTRIBUTES.keys())
        if fields:
            fields = set(fields.split(','))
            attributes = set(self.COMPLEX_ATTRIBUTES).intersection(fields)
            for attribute in attributes:
                add_fields = set([key for key in self.FLATTEN_ATTRIBUTES.
                    keys() if key.startswith(attribute)])
                fields = fields.union(add_fields)

            include_fields = set(
                _vnf_lcm_op_occs.VnfLcmOpOcc.SIMPLE_ATTRIBUTES).union(fields)
        elif exclude_default:
            include_fields = set(
                _vnf_lcm_op_occs.VnfLcmOpOcc.SIMPLE_ATTRIBUTES)
        elif exclude_fields:
            exclude_fields = set(exclude_fields.split(','))
            exclude_additional_attributes = set(
                self.COMPLEX_ATTRIBUTES).intersection(exclude_fields)
            for attribute in exclude_additional_attributes:
                fields = set([key for key in self.FLATTEN_ATTRIBUTES.keys()
                    if key.startswith(attribute)])
                exclude_fields = exclude_fields.union(fields)

            include_fields = set(self.FLATTEN_ATTRIBUTES.keys()) - \
                exclude_fields

        return [
            self._get_vnf_lcm_op_occs_list(
                vnf_lcm_op_occ, include_fields=include_fields)
            for vnf_lcm_op_occ in vnf_lcm_op_occs]



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import netaddr

from oslo_config import cfg
from oslo_log import log as logging
import random
import string
from tacker._i18n import _
from tacker.common import utils
from tacker.vnfm.monitor_drivers import abstract_driver


LOG = logging.getLogger(__name__)

OPTS = [
    cfg.HostAddressOpt('host', default=utils.get_hostname(),
                       help=_('Address which drivers use to trigger')),
    cfg.PortOpt('port', default=9890,
               help=_('port number which drivers use to trigger'))
]
cfg.CONF.register_opts(OPTS, group='ceilometer')


def config_opts():
    return [('ceilometer', OPTS)]


ALARM_INFO = (
    ALARM_ACTIONS, OK_ACTIONS, REPEAT_ACTIONS, ALARM,
    INSUFFICIENT_DATA_ACTIONS, DESCRIPTION, ENABLED, TIME_CONSTRAINTS,
    SEVERITY,
) = ('alarm_actions', 'ok_actions', 'repeat_actions', 'alarm',
     'insufficient_data_actions', 'description', 'enabled', 'time_constraints',
     'severity',
     )


class VNFMonitorCeilometer(
        abstract_driver.VNFMonitorAbstractDriver):
    def get_type(self):
        return 'ceilometer'

    def get_name(self):
        return 'ceilometer'

    def get_description(self):
        return 'Tacker VNFMonitor Ceilometer Driver'

    def _create_alarm_url(self, vnf_id, mon_policy_name, mon_policy_action):
        # alarm_url = 'http://host:port/v1.0/vnfs/vnf-uuid/monitoring-policy
        # -name/action-name?key=8785'
        host = cfg.CONF.ceilometer.host
        port = cfg.CONF.ceilometer.port
        LOG.info("Tacker in heat listening on %(host)s:%(port)s",
                 {'host': host,
                  'port': port})
        origin = "http://%(host)s:%(port)s/v1.0/vnfs" % {
            'host': host, 'port': port}
        if netaddr.valid_ipv6(host):
            origin = "http://[%(host)s]:%(port)s/v1.0/vnfs" % {
                'host': host, 'port': port}
        access_key = ''.join(
            random.SystemRandom().choice(
                string.ascii_lowercase + string.digits)
            for _ in range(8))
        alarm_url = "".join([origin, '/', vnf_id, '/', mon_policy_name, '/',
                             mon_policy_action, '/', access_key])
        return alarm_url

    def call_alarm_url(self, vnf, kwargs):
        '''must be used after call heat-create in plugin'''
        return self._create_alarm_url(**kwargs)

    def _process_alarm(self, alarm_id, status):
        if alarm_id and status == ALARM:
            return True

    def process_alarm(self, vnf, kwargs):
        '''Check alarm state. if available, will be processed'''
        return self._process_alarm(**kwargs)

    def monitor_url(self, plugin, context, vnf):
        pass

    def monitor_call(self, vnf, kwargs):
        pass



# Copyright 2011, VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Borrowed from nova code base, more utilities will be added/borrowed as and
# when needed.

"""Utilities and helper functions."""

import functools
from functools import reduce
import inspect
import logging as std_logging
import math
import os
import random
import re
import signal
import socket
from stevedore import driver
import string
import sys

from eventlet.green import subprocess
import netaddr
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
import urllib
from urllib import parse as urlparse
try:
    from eventlet import sleep
except ImportError:
    from time import sleep

from tacker._i18n import _
from tacker.common import constants as q_const
from tacker.common import exceptions
from tacker.common import safe_utils


TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
LOG = logging.getLogger(__name__)
SYNCHRONIZED_PREFIX = 'tacker-'
MEM_UNITS = {
    "MB": {
        "MB": {
            "op": "*",
            "val": "1"
        },
        "GB": {
            "op": "/",
            "val": "1024"
        }
    },
    "GB": {
        "MB": {
            "op": "*",
            "val": "1024"
        },
        "GB": {
            "op": "*",
            "val": "1"
        }
    }
}
CONF = cfg.CONF
synchronized = lockutils.synchronized_with_prefix(SYNCHRONIZED_PREFIX)
MAX_COOP_READER_BUFFER_SIZE = 134217728

if hasattr(inspect, 'getfullargspec'):
    getargspec = inspect.getfullargspec
else:
    getargspec = inspect.getargspec


def find_config_file(options, config_file):
    """Return the first config file found.

    We search for the paste config file in the following order:
    * If --config-file option is used, use that
    * Search for the configuration files via common cfg directories
    :retval Full path to config file, or None if no config file found
    """
    def fix_path(p):
        return os.path.abspath(os.path.expanduser(p))
    if options.get('config_file'):
        if os.path.exists(options['config_file']):
            return fix_path(options['config_file'])

    dir_to_common = os.path.dirname(os.path.abspath(__file__))
    root = os.path.join(dir_to_common, '..', '..', '..', '..')
    # Handle standard directory search for the config file
    config_file_dirs = [fix_path(os.path.join(os.getcwd(), 'etc')),
                        fix_path(os.path.join('~', '.tacker-venv', 'etc',
                                              'tacker')),
                        fix_path('~'),
                        os.path.join(cfg.CONF.state_path, 'etc'),
                        os.path.join(cfg.CONF.state_path, 'etc', 'tacker'),
                        fix_path(os.path.join('~', '.local',
                                              'etc', 'tacker')),
                        '/usr/etc/tacker',
                        '/usr/local/etc/tacker',
                        '/etc/tacker/',
                        '/etc']

    if 'plugin' in options:
        config_file_dirs = [
            os.path.join(x, 'tacker', 'plugins', options['plugin'])
            for x in config_file_dirs
        ]

    if os.path.exists(os.path.join(root, 'plugins')):
        plugins = [fix_path(os.path.join(root, 'plugins', p, 'etc'))
                   for p in os.listdir(os.path.join(root, 'plugins'))]
        plugins = [p for p in plugins if os.path.isdir(p)]
        config_file_dirs.extend(plugins)

    for cfg_dir in config_file_dirs:
        cfg_file = os.path.join(cfg_dir, config_file)
        if os.path.exists(cfg_file):
            return cfg_file


def _subprocess_setup():
    # Python installs a SIGPIPE handler by default. This is usually not what
    # non-Python subprocesses expect.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def subprocess_popen(args, stdin=None, stdout=None, stderr=None, shell=False,
                     env=None):
    return subprocess.Popen(args, shell=shell, stdin=stdin, stdout=stdout,
                            stderr=stderr, preexec_fn=_subprocess_setup,
                            close_fds=True, env=env)


def get_hostname():
    return socket.gethostname()


def dict2tuple(d):
    items = list(d.items())
    items.sort()
    return tuple(items)


def log_opt_values(log):
    cfg.CONF.log_opt_values(log, std_logging.DEBUG)


def is_valid_vlan_tag(vlan):
    return q_const.MIN_VLAN_TAG <= vlan <= q_const.MAX_VLAN_TAG


def is_valid_ipv4(address):
    """Verify that address represents a valid IPv4 address."""
    try:
        return netaddr.valid_ipv4(address)
    except Exception:
        return False


def change_memory_unit(mem, to):
    """Change the memory value(mem) based on the unit('to') specified.

    If the unit is not specified in 'mem', by default, it is considered
    as "MB". And this method returns only integer.
    """
    mem = str(mem) + " MB" if str(mem).isdigit() else mem.upper()
    for unit, value in (MEM_UNITS).items():
        mem_arr = mem.split(unit)
        if len(mem_arr) < 2:
            continue
        return eval(mem_arr[0] +
                    MEM_UNITS[unit][to]["op"] +
                    MEM_UNITS[unit][to]["val"])


def load_class_by_alias_or_classname(namespace, name):
    """Load class using stevedore alias or the class name.

    Load class using the stevedore driver manager
    :param namespace: namespace where the alias is defined
    :param name: alias or class name of the class to be loaded
    :returns: class if calls can be loaded
    :raises ImportError: if class cannot be loaded
    """
    if not name:
        LOG.error("Alias or class name is not set")
        raise ImportError(_("Class not found."))
    try:
        # Try to resolve class by alias
        mgr = driver.DriverManager(namespace, name)
        class_to_load = mgr.driver
    except RuntimeError:
        e1_info = sys.exc_info()
        # Fallback to class name
        try:
            class_to_load = importutils.import_class(name)
        except (ImportError, ValueError):
            LOG.error("Error loading class by alias",
                      exc_info=e1_info)
            LOG.error("Error loading class by class name",
                      exc_info=True)
            raise ImportError(_("Class not found."))
    return class_to_load


def deep_update(orig_dict, new_dict):
    for key, value in new_dict.items():
        if isinstance(value, dict):
            if key in orig_dict and isinstance(orig_dict[key], dict):
                deep_update(orig_dict[key], value)
                continue

        orig_dict[key] = value


def generate_resource_name(resource, prefix='tmpl'):
    return prefix + '-' \
        + ''.join(random.SystemRandom().choice(
                  string.ascii_lowercase + string.digits)
          for _ in range(16)) \
        + '-' + resource


def get_auth_url_v3(auth_url):
    if re.match('.+v3/?$', auth_url) is not None:
        return auth_url
    else:
        return '{0}/v3'.format(auth_url)


def none_from_string(orig_str):
    none_values = ['', 'None', 'NONE', 'null', 'NULL']
    if orig_str in none_values:
        return None
    else:
        return orig_str


def expects_func_args(*args):
    def _decorator_checker(dec):
        @functools.wraps(dec)
        def _decorator(f):
            base_f = safe_utils.get_wrapped_function(f)
            argspec = getargspec(base_f)
            if argspec[1] or argspec[2] or set(args) <= set(argspec[0]):
                # NOTE (nirajsingh): We can't really tell if correct stuff will
                # be passed if it's a function with *args or **kwargs so
                # we still carry on and hope for the best
                return dec(f)
            else:
                raise TypeError("Decorated function %(f_name)s does not "
                                "have the arguments expected by the "
                                "decorator %(d_name)s" %
                                {'f_name': base_f.__name__,
                                 'd_name': dec.__name__})

        return _decorator

    return _decorator_checker


def cooperative_iter(iter):
    """Prevent eventlet thread starvation during iteration

    Return an iterator which schedules after each
    iteration. This can prevent eventlet thread starvation.

    :param iter: an iterator to wrap
    """
    try:
        for chunk in iter:
            sleep(0)
            yield chunk
    except Exception as err:
        with excutils.save_and_reraise_exception():
            msg = _("Error: cooperative_iter exception %s") % err
            LOG.error(msg)


def cooperative_read(fd):
    """Prevent eventlet thread starvationafter each read operation.

    Wrap a file descriptor's read with a partial function which schedules
    after each read. This can prevent eventlet thread starvation.

    :param fd: a file descriptor to wrap
    """
    def readfn(*args):
        result = fd.read(*args)
        sleep(0)
        return result
    return readfn


def chunkreadable(iter, chunk_size=65536):
    """Wrap a readable iterator.

    Wrap a readable iterator with a reader yielding chunks of
    a preferred size, otherwise leave iterator unchanged.

    :param iter: an iter which may also be readable
    :param chunk_size: maximum size of chunk
    """
    return chunkiter(iter, chunk_size) if hasattr(iter, 'read') else iter


def chunkiter(fp, chunk_size=65536):
    """Convert iterator to a file-like object.

    Return an iterator to a file-like obj which yields fixed size chunks

    :param fp: a file-like object
    :param chunk_size: maximum size of chunk
    """
    while True:
        chunk = fp.read(chunk_size)
        if chunk:
            yield chunk
        else:
            break


# TODO(esto.aln): Consider to move this function to
# convert_camelcase_to_snakecase(). We will consider the correct approach
# to modify the common function so as not to introduce degrade.
def convert_string_to_snakecase(name):
    """Converts a string from camelCase to snake_case."""
    name_with_underscores = re.sub(
        '(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2',
                  name_with_underscores).lower()


def convert_camelcase_to_snakecase(request_data):
    """Converts dict keys or list of dict keys from camelCase to snake_case.

    Returns a dict with keys or list with dict keys, in snake_case.
    This method takes care only keys in a `dict` or `dicts in a list`.
    For simple list with string items, the elements which are actual values
    are ignored during conversion.

    :param request_data: dict with keys or list with items, in camelCase.
    """
    if isinstance(request_data, dict):
        new_dict = {}
        for key, property_value in request_data.items():
            property_value = convert_camelcase_to_snakecase(property_value)
            underscore_joined = convert_string_to_snakecase(key)
            new_dict[underscore_joined] = property_value
        return new_dict

    if isinstance(request_data, list):
        new_list = []
        for property_value in request_data:
            new_list.append(
                convert_camelcase_to_snakecase(property_value))
        return new_list

    return request_data


def convert_snakecase_to_camelcase(request_data):
    """Converts dict keys or list of dict keys from snake_case to camelCase.

    Returns a dict with keys or list with dict key, in camelCase.
    This method takes care only keys in a `dict` or `dicts in a list`.
    For simple list with string items, the elements which are actual values
    are ignored during conversion.
    Also, Snake case is a notation method that uses underscores to connect
    words. For that reason, if the initial word of the key in dict starts
    with '_', this function ignore to convert the key.

    :param request_data: dict with keys or list with items, in snake_case.

    Example:
        Before::

            [
                {"vnf_lcm_op_occ_id" : "uuid"},
                {
                    "_links" : {
                        "vnf_lcm_op_occ": {"href": "resource_link"}
                    }
                }
            ]

        After::

            [
                {"vnfLcmOpOccId": "uuid"},
                {
                    "_links": {
                        "vnfLcmOpOcc": {"href": "resource_link"}
                    }
                }
            ]
    """
    def convert(name):
        return re.sub('_([a-z])',
                      lambda match: match.group(1).upper(), name)

    if isinstance(request_data, dict):
        new_dict = {}
        for key, property_value in request_data.items():
            property_value = convert_snakecase_to_camelcase(property_value)
            camelcase = key if key.startswith('_') else convert(key)
            new_dict[camelcase] = property_value
        return new_dict

    if isinstance(request_data, list):
        new_list = []
        for property_value in request_data:
            new_list.append(
                convert_snakecase_to_camelcase(property_value))
        return new_list

    return request_data


def is_url(url):
    try:
        urllib.request.urlopen(url)
        return True
    except Exception:
        return False


def flatten_dict(data, prefix=''):
    ret = {}
    for key, val in data.items():
        key = prefix + key
        if isinstance(val, dict):
            ret.update(flatten_dict(val, key + '/'))
        else:
            ret[key] = val
    return ret


def deepgetattr(obj, attr):
    """Recurses through an attribute chain to get the ultimate value."""
    return reduce(getattr, attr.split('.'), obj)


def is_valid_url(url):
    url_parts = urlparse.urlparse(url)

    if not (url_parts.scheme and url_parts.netloc and url_parts.path):
        return False

    schemes = ['http', 'https', 'ftp']
    if url_parts.scheme not in schemes:
        return False

    return True


class CooperativeReader(object):
    """An eventlet thread friendly class for reading in image data.

    When accessing data either through the iterator or the read method
    we perform a sleep to allow a co-operative yield. When there is more than
    one image being uploaded/downloaded this prevents eventlet thread
    starvation, ie allows all threads to be scheduled periodically rather than
    having the same thread be continuously active.
    """
    def __init__(self, fd):
        """Construct an CooperativeReader object.

        :param fd: Underlying image file object

        """
        self.fd = fd
        self.iterator = None
        # NOTE(nirajsingh): if the underlying supports read(), overwrite the
        # default iterator-based implementation with cooperative_read which
        # is more straightforward
        if hasattr(fd, 'read'):
            self.read = cooperative_read(fd)
        else:
            self.iterator = None
            self.buffer = b''
            self.position = 0

    def read(self, length=None):
        """Return the requested amount of bytes.

        Fetching the next chunk of the underlying iterator when needed.
        This is replaced with cooperative_read in __init__ if the underlying
        fd already supports read().

        """

        if length is None:
            if len(self.buffer) - self.position > 0:
                # if no length specified but some data exists in buffer,
                # return that data and clear the buffer
                result = self.buffer[self.position:]
                self.buffer = b''
                self.position = 0
                return bytes(result)
            else:
                # otherwise read the next chunk from the underlying iterator
                # and return it as a whole. Reset the buffer, as subsequent
                # calls may specify the length
                try:
                    if self.iterator is None:
                        self.iterator = self.__iter__()
                    return next(self.iterator)
                except StopIteration:
                    return b''
                finally:
                    self.buffer = b''
                    self.position = 0
        else:
            result = bytearray()
            while len(result) < length:
                if self.position < len(self.buffer):
                    to_read = length - len(result)
                    chunk = self.buffer[self.position:self.position + to_read]
                    result.extend(chunk)

                    # This check is here to prevent potential OOM issues if
                    # this code is called with unreasonably high values of read
                    # size. Currently it is only called from the HTTP clients
                    # of Glance backend stores, which use httplib for data
                    # streaming, which has readsize hardcoded to 8K, so this
                    # check should never fire. Regardless it still worths to
                    # make the check, as the code may be reused somewhere else.
                    if len(result) >= MAX_COOP_READER_BUFFER_SIZE:
                        raise exceptions.LimitExceeded()
                    self.position += len(chunk)
                else:
                    try:
                        if self.iterator is None:
                            self.iterator = self.__iter__()
                        self.buffer = next(self.iterator)
                        self.position = 0
                    except StopIteration:
                        self.buffer = b''
                        self.position = 0
                        return bytes(result)
            return bytes(result)

    def __iter__(self):
        return cooperative_iter(self.fd.__iter__())


class LimitingReader(object):
    """Limit Reader to read data past to configured allowed amount.

    Reader designed to fail when reading image data past the configured
    allowable amount.
    """
    def __init__(self, data, limit,
                 exception_class=exceptions.CSARFileSizeLimitExceeded):
        """Construct an LimitingReader object.

        :param data: Underlying image data object
        :param limit: maximum number of bytes the reader should allow
        :param exception_class: Type of exception to be raised
        """
        self.data = data
        self.limit = limit
        self.bytes_read = 0
        self.exception_class = exception_class

    def __iter__(self):
        for chunk in self.data:
            self.bytes_read += len(chunk)
            if self.bytes_read > self.limit:
                raise self.exception_class()
            else:
                yield chunk

    def read(self, i):
        result = self.data.read(i)
        self.bytes_read += len(result)
        if self.bytes_read > self.limit:
            raise self.exception_class()
        return result


class MemoryUnit(object):

    UNIT_SIZE_DEFAULT = 'B'
    UNIT_SIZE_DICT = {'B': 1, 'kB': 1000, 'KiB': 1024, 'MB': 1000000,
                      'MiB': 1048576, 'GB': 1000000000,
                      'GiB': 1073741824, 'TB': 1000000000000,
                      'TiB': 1099511627776}

    @staticmethod
    def convert_unit_size_to_num(size, unit=None):
        """Convert given size to a number representing given unit.

        If unit is None, convert to a number representing UNIT_SIZE_DEFAULT
        :param size: unit size e.g. 1 TB
        :param unit: unit to be converted to e.g GB
        :return: converted number e.g. 1000 for 1 TB size and unit GB
        """
        if unit:
            unit = MemoryUnit.validate_unit(unit)
        else:
            unit = MemoryUnit.UNIT_SIZE_DEFAULT
            LOG.info(_('A memory unit is not provided for size; using the '
                       'default unit %(default)s.') % {'default': 'B'})
        result = re.sub(r'\s+', ' ', size).split(' ')
        if len(result) == 2:
            if result[1]:
                unit_size = MemoryUnit.validate_unit(result[1])
                converted = int(str_to_num(result[0]) *
                                MemoryUnit.UNIT_SIZE_DICT[unit_size] *
                                math.pow(MemoryUnit.UNIT_SIZE_DICT
                                         [unit], -1))
                LOG.info(_('Given size %(size)s is converted to %(num)s '
                           '%(unit)s.') % {'size': size,
                         'num': converted, 'unit': unit})
            else:
                msg = _('Size is not given for software image data.')
                LOG.error(msg)
                raise ValueError(msg)
        else:
            msg = _('Error while converting unit "{0}" to number.'
                    ).format(size)
            LOG.error(msg)
            raise ValueError(msg)
        return converted

    @staticmethod
    def validate_unit(unit):
        if unit in MemoryUnit.UNIT_SIZE_DICT.keys():
            return unit
        else:
            for key in MemoryUnit.UNIT_SIZE_DICT.keys():
                if key.upper() == unit.upper():
                    return key

            msg = _('Provided unit "{0}" is not valid. The valid units are '
                    '{1}').format(unit, MemoryUnit.UNIT_SIZE_DICT.keys())
            LOG.error(msg)
            raise ValueError(msg)


def str_to_num(value):
    """Convert a string representation of a number into a numeric type."""
    if (isinstance(value, int) or
            isinstance(value, float)):
        return value

    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return None


def str_to_bytes(value):
    """Convert string to bytes"""
    if isinstance(value, str):
        value = bytes(value, 'utf-8')
    elif value is not None:
        value = bytes(value)

    return value



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add unique constraint on deleted_at

Revision ID: e7993093baf1
Revises: c256228ed37c
Create Date: 2017-04-19 10:57:22.157326

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'e7993093baf1'
down_revision = 'c256228ed37c'

from alembic import op


def _drop_unique_constraint(table):
    op.drop_constraint(
        constraint_name='uniq_%s0tenant_id0name' % table,
        table_name=table, type_='unique')


def _add_unique_constraint(table):
    op.create_unique_constraint(
        constraint_name='uniq_%s0tenant_id0name0deleted_at' % table,
        table_name=table,
        columns=['tenant_id', 'name', 'deleted_at'])


def upgrade(active_plugins=None, options=None):
    for table in ['vnf', 'vnfd', 'vims', 'ns', 'nsd']:
        _drop_unique_constraint(table)
        _add_unique_constraint(table)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import yaml

from oslo_config import cfg

from tacker.common import cmd_executer
from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file


CONF = cfg.CONF
VNF_CIRROS_CREATE_TIMEOUT = 120


class TestRemoteCommandExecutor(base.BaseTackerTest):

    def _test_create_vnf(self, vnfd_file, vnf_name):
        input_yaml = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)

        # create vnf directly from template
        vnf_arg = {'vnf': {'vnfd_template': tosca_dict, 'name': vnf_name}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        vnf_show_out = self.client.show_vnf(vnf_id)['vnf']
        self.assertIsNotNone(vnf_show_out['mgmt_ip_address'])

        # Fetch mgmt ip of VNF
        mgmt_ip = eval(vnf_show_out['mgmt_ip_address'])['VDU1']

        return vnf_id, mgmt_ip

    def _test_cmd_executor(self, vnfd_file, vnf_name):
        vnf_id, mgmt_ip = self._test_create_vnf(vnfd_file, vnf_name)

        # Wait for VM initialization. It takes time after VM started until
        # sshd starts. It is a bit ad-hoc but adopted it for simplicity.
        time.sleep(100)

        # Login on VNF instance, and execute 'hostname' command to verify
        # connection and command output.
        usr = 'cirros'
        psswrd = 'gocubsgo'
        cmd = 'hostname'
        rcmd_executor = cmd_executer.RemoteCommandExecutor(user=usr,
                                                           password=psswrd,
                                                           host=mgmt_ip)
        result = rcmd_executor.execute_command(cmd)
        self.assertEqual(cmd, result.get_command())
        self.assertEqual(0, result.get_return_code())
        self.assertIn('test-vdu', result.get_stdout()[0])

        self._test_delete_vnf(vnf_id)

    def _test_delete_vnf(self, vnf_id):
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

    def test_cmd_executor(self):
        self._test_cmd_executor('sample-tosca-vnfd.yaml',
                                'test_tosca_vnf_with_cirros_inline')



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
from unittest import mock

from oslo_config import cfg

from tacker import context as t_context
from tacker.nfvo.drivers.vim import kubernetes_driver
from tacker.tests.unit import base


OPTS = [cfg.StrOpt('user_domain_id',
                   default='default',
                   help='User Domain Id'),
        cfg.StrOpt('project_domain_id',
                   default='default',
                   help='Project Domain Id'),
        cfg.StrOpt('password',
                   default='default',
                   help='User Password'),
        cfg.StrOpt('username',
                   default='default',
                   help='User Name'),
        cfg.StrOpt('user_domain_name',
                   default='default',
                   help='Use Domain Name'),
        cfg.StrOpt('project_name',
                   default='default',
                   help='Project Name'),
        cfg.StrOpt('project_domain_name',
                   default='default',
                   help='Project Domain Name'),
        cfg.StrOpt('auth_url',
                   default='http://localhost/identity/v3',
                   help='Keystone endpoint')]

cfg.CONF.register_opts(OPTS, 'keystone_authtoken')
CONF = cfg.CONF


class FakeKubernetesAPI(mock.Mock):
    pass


class FakeKeymgrAPI(mock.Mock):
    pass


class mock_dict(dict):
    def __getattr__(self, item):
        return self.get(item)
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class TestKubernetes_Driver(base.TestCase):
    def setUp(self):
        super(TestKubernetes_Driver, self).setUp()
        self._mock_kubernetes()
        self.config_fixture.config(group='k8s_vim', use_barbican=True)
        self.kubernetes_driver = kubernetes_driver.Kubernetes_Driver()
        self.vim_obj = self.get_vim_obj()
        self.addCleanup(mock.patch.stopall)
        self._mock_keymgr()

    def _mock_kubernetes(self):
        self.kubernetes_api = mock.Mock(wraps=FakeKubernetesAPI())
        fake_kubernetes_api = mock.Mock()
        fake_kubernetes_api.return_value = self.kubernetes_api
        self._mock('tacker.common.container.kubernetes_utils.'
                   'KubernetesHTTPAPI', fake_kubernetes_api)

    def _mock_keymgr(self):
        self.keymgr = mock.Mock(wraps=FakeKeymgrAPI())
        fake_keymgr = mock.Mock()
        fake_keymgr.return_value = self.keymgr
        self._mock(
            'tacker.keymgr.barbican_key_manager.BarbicanKeyManager',
            fake_keymgr)

    def get_vim_obj(self):
        return {'id': '647a91c3-d436-43e6-a1e8-71118dde84ce',
                'type': 'kubernetes',
                'auth_url': 'https://localhost:6443',
                'auth_cred': {'username': 'test_user',
                              'password': 'test_password',
                              'ssl_ca_cert': 'None'},
                'name': 'vim-kubernetes',
                'vim_project': {'name': 'default'}}

    def get_vim_obj_barbican(self):
        return {'id': '647a91c3-d436-43e6-a1e8-71118dde84ce',
                'type': 'kubernetes',
                'auth_url': 'https://localhost:6443',
                'auth_cred': {'username': 'test_user',
                              'password': 'test_password',
                              'ssl_ca_cert': 'abcxyz',
                              'key_type': 'barbican_key',
                              'secret_uuid': 'fake-secret-uuid'},
                'name': 'vim-kubernetes',
                'vim_project': {'name': 'default'}}

    def test_register_k8sclient(self):
        dict = {'name': 'default'}
        name = namedtuple("name", dict.keys())(*dict.values())
        dict = {'metadata': name}
        metadata = namedtuple("metadata", dict.keys())(*dict.values())
        dict = {'items': [metadata]}
        namespaces = namedtuple("namespace", dict.keys())(*dict.values())
        attrs = {'list_namespace.return_value': namespaces}
        mock_k8s_client = mock.Mock()
        mock_k8s_coreV1Client = mock.Mock(**attrs)
        auth_obj = {'username': 'test_user',
                    'password': 'test_password',
                    'ssl_ca_cert': 'None',
                    'auth_url': 'https://localhost:6443'}
        self._test_register_vim(self.vim_obj, mock_k8s_client,
                                mock_k8s_coreV1Client)
        mock_k8s_coreV1Client.list_namespace.assert_called_once_with()
        self.kubernetes_api. \
            get_core_api_client.assert_called_once_with(auth_obj)

    def _test_register_vim(self, vim_obj, mock_k8s_client,
                           mock_k8s_coreV1Client):
        self.kubernetes_api. \
            get_core_api_client.return_value = mock_k8s_client
        self.kubernetes_api. \
            get_core_v1_api_client.return_value = mock_k8s_coreV1Client
        fernet_attrs = {'encrypt.return_value': 'encrypted_password'}
        mock_fernet_obj = mock.Mock(**fernet_attrs)
        mock_fernet_key = 'test_fernet_key'
        self.kubernetes_api.create_fernet_key.return_value = (mock_fernet_key,
                                                              mock_fernet_obj)
        self.kubernetes_api.create_ca_cert_tmp_file.\
            return_value = ('file_descriptor', 'file_path')
        self.kubernetes_driver.register_vim(vim_obj)
        mock_fernet_obj.encrypt.assert_called_once_with(mock.ANY)

    def test_deregister_vim_barbican(self):
        self.keymgr.delete.return_value = None
        vim_obj = self.get_vim_obj_barbican()
        self.kubernetes_driver.deregister_vim(vim_obj)
        self.keymgr.delete.assert_called_once_with(
            t_context.generate_tacker_service_context(), 'fake-secret-uuid')

    def test_encode_vim_auth_barbican(self):
        self.config_fixture.config(group='k8s_vim',
                                   use_barbican=True)
        fernet_attrs = {'encrypt.return_value': 'encrypted_password'}
        mock_fernet_obj = mock.Mock(**fernet_attrs)
        mock_fernet_key = 'test_fernet_key'
        self.keymgr.store.return_value = 'fake-secret-uuid'
        self.kubernetes_api.create_fernet_key.return_value = (mock_fernet_key,
                                                              mock_fernet_obj)

        vim_obj = self.get_vim_obj()
        self.kubernetes_driver.encode_vim_auth(
            vim_obj['id'], vim_obj['auth_cred'])

        self.keymgr.store.assert_called_once_with(
            t_context.generate_tacker_service_context(), 'test_fernet_key')
        mock_fernet_obj.encrypt.assert_called_once_with(mock.ANY)
        self.assertEqual(vim_obj['auth_cred']['key_type'],
                         'barbican_key')
        self.assertEqual(vim_obj['auth_cred']['secret_uuid'],
                         'fake-secret-uuid')



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Key manager API
"""

import abc


class KeyManager(object, metaclass=abc.ABCMeta):
    """Base Key Manager Interface

    A Key Manager is responsible for creating, reading, and deleting keys.
    """

    @abc.abstractmethod
    def __init__(self, auth_url):
        """Instantiate a KeyManager object.

        Creates a KeyManager object with implementation specific details
        obtained from the supplied configuration.
        """
        pass

    @abc.abstractmethod
    def store(self, context, managed_object, expiration=None):
        """Stores a managed object with the key manager.

        This method stores the specified managed object and returns its UUID
        that identifies it within the key manager. If the specified context
        does not permit the creation of keys, then a NotAuthorized exception
        should be raised.
        """
        pass

    @abc.abstractmethod
    def get(self, context, managed_object_id, metadata_only=False):
        """Retrieves the specified managed object.

        Implementations should verify that the caller has permissions to
        retrieve the managed object by checking the context object passed in
        as context. If the user lacks permission then a NotAuthorized
        exception is raised.

        If the caller requests only metadata, then the object that is
        returned will contain only the secret metadata and no secret bytes.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling
        this method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.
        """
        pass

    @abc.abstractmethod
    def delete(self, context, managed_object_id):
        """Deletes the specified managed object.

        Implementations should verify that the caller has permission to delete
        the managed object by checking the context object (context). A
        NotAuthorized exception should be raised if the caller lacks
        permission.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling this
        method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.
        """
        pass



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# key
KEY_ACTION = 'action'
KEY_KWARGS = 'kwargs'
KEY_SCALE = 'scale_stack_id'

# ACTION type
ACTION_CREATE_VNF = 'create_vnf'
ACTION_UPDATE_VNF = 'update_vnf'
ACTION_DELETE_VNF = 'delete_vnf'
ACTION_INSTANTIATE_VNF = 'instantiate_vnf'
ACTION_TERMINATE_VNF = 'terminate_vnf'
ACTION_HEAL_VNF = 'heal_vnf'
ACTION_SCALE_IN_VNF = 'scale_in_vnf'
ACTION_SCALE_OUT_VNF = 'scale_out_vnf'
